/* Copyright (C) 2012 Oleg Moskalenko oleg.moskalenko@citrix.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/limits.h>

#include <errno.h>
#include <err.h>
#include <langinfo.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <wctype.h>

#include "sort.h"

/* counterparts of wcs functions */

/*
 * Allocate a new binary string of specified size
 */
bwstring* bwsalloc(size_t sz)
{
	bwstring *ret = (bwstring*)sort_malloc(sizeof(bwstring)+SIZEOF_WCHAR_STRING(sz));
	if(!ret)
		err(2,NULL);
	else {
		ret->sz=sz;
		ret->len=0;
		ret->str[ret->len]=L'\0';
		return ret;
	}
}

/*
 * Re-allocate the binary string, preserving the content when possible
 */
bwstring* bwsrealloc(bwstring* s, size_t sz)
{
	bool empty = (s == NULL);
	bwstring *ret = (bwstring*)sort_realloc(s, sizeof(bwstring)+SIZEOF_WCHAR_STRING(sz));
	if(!ret)
		err(2,NULL);
	else {
		ret->sz=sz;
		if(!empty) {
			if(ret->len>sz) {
				ret->len=sz;
			}
		} else {
			ret->len=0;
		}
		ret->str[ret->len]=L'\0';
		return ret;
	}
}

/*
 * Create a copy of binary string.
 * New string size equals the length of the old string.
 */
bwstring* bwsdup(const bwstring *s)
{
	if(!s) return NULL;
	else {
		bwstring* ret=bwsalloc(s->len);
		ret->len=s->len;
		memcpy(ret->str,s->str,SIZEOF_WCHAR_STRING(s->len));
		ret->str[ret->len]=L'\0';
		return ret;
	}
}

/*
 * Create a new binary string from a C string.
 */
bwstring* bwssdup(const wchar_t *str)
{
	if(!str) return NULL;
	else {
		size_t len=wcslen(str);
		bwstring* ret=bwsalloc(len);
		ret->len=len;
		wcscpy(ret->str,str);
		ret->str[ret->len]=L'\0';
		return ret;
	}
}

/*
 * Create a new binary string from a raw binary buffer.
 */
bwstring* bwssbdup(const wchar_t *str, size_t len)
{
	if (!str)
		return NULL;
	else {
		bwstring* ret = bwsalloc(len);
		ret->len = len;
		memcpy(ret->str, str,SIZEOF_WCHAR_STRING(len));
		ret->str[ret->len] = L'\0';
		return ret;
	}
}

/*
 * De-allocate object memory
 */
void bwsfree(bwstring* s)
{
	if(s)
		sort_free(s);
}

/*
 * Copy content of src binary string to dst.
 * If the capacity of the dst string is not sufficient,
 * then the data is truncated.
 */
size_t bwscpy(bwstring *dst, const bwstring *src)
{
	size_t ret=src->len;
	size_t nums=ret;
	if(nums>dst->sz)
		nums=dst->sz;
	dst->len=nums;
	memcpy(dst->str,src->str,SIZEOF_WCHAR_STRING(nums));
	dst->str[dst->len]=L'\0';
	return ret;
}

/*
 * Copy content of src binary string to dst,
 * with specified number of symbols to be copied.
 * If the capacity of the dst string is not sufficient,
 * then the data is truncated.
 */
bwstring* bwsncpy(bwstring *dst, const bwstring *src, size_t size)
{
	size_t nums=src->len;
	if(nums>dst->sz)
		nums=dst->sz;
	if(nums>size)
		nums=size;
	dst->len=nums;
	memcpy(dst->str,src->str,SIZEOF_WCHAR_STRING(nums));
	dst->str[dst->len]=L'\0';
	return dst;
}

/*
 * Copy content of src binary string to dst,
 * with specified number of symbols to be copied.
 * An offset value can be specified, from the start of src string.
 * If the capacity of the dst string is not sufficient,
 * then the data is truncated.
 */
bwstring* bwsnocpy(bwstring *dst, const bwstring *src, size_t offset, size_t size)
{
	if(offset>=src->len) {
		if(dst->sz>0) {
			dst->str[0]=0;
			dst->len=0;
		}
	} else {
		size_t nums=src->len-offset;
		if(nums>dst->sz)
			nums=dst->sz;
		if(nums>size)
			nums=size;
		dst->len=nums;
		memcpy(dst->str,src->str+offset,SIZEOF_WCHAR_STRING(nums));
		dst->str[dst->len]=L'\0';
	}
	return dst;
}

/*
 * Compare two binary strings, according to the locale setting.
 */
int bwscoll_bin(const bwstring* bws1, const bwstring *bws2)
{
	int res = 0;

	const size_t len = 128;
	static wchar_t s1[129];
	static wchar_t s2[129];

	wchar_t c1 = 0, c2 = 0;
	size_t i = 0;
	size_t maxlen = bws1->len;

	if (maxlen > bws2->len)
		maxlen = bws2->len;

	while (i < maxlen) {
		size_t j = 0;
		res = 0;
		for (j = 0; ((j < len) && (i < maxlen)); ++j, ++i) {
			c1 = bws1->str[i];
			c2 = bws2->str[i];
			if (!c1 || !c2)
				break;
			s1[j] = c1;
			s2[j] = c2;
		}
		s1[j] = L'\0';
		s2[j] = L'\0';
		res = wcscoll(s1, s2);
		if (res)
			return res;
		if (!c1 && c2)
			return -1;
		else if (!c2 && c1)
			return +1;
		else if (!c1 && !c2)
			++i;
	}

	if (bws1->len < bws2->len)
		return -1;
	else if (bws1->len > bws2->len)
		return +1;

	return 0;
}

/*
 * Compare two binary strings, according to the locale setting.
 * Ignore the difference between upper and lower case symbols.
 */
int bwscasecoll(const bwstring* bws1, const bwstring *bws2)
{
	const size_t len=128;
	static wchar_t s1[129];
	static wchar_t s2[129];

	wchar_t c1=0, c2=0;
	size_t i=0;
	size_t maxlen=bws1->len;
	if(maxlen>bws2->len)
		maxlen=bws2->len;

	while(i<maxlen) {
		size_t j=0;
		int res=0;
		for(j=0;((j<len)&&(i<maxlen));++j,++i) {
			c1=bws1->str[i];
			c2=bws2->str[i];
			if(!c1 || !c2) break;
			s1[j]=towlower(c1);
			s2[j]=towlower(c2);
		}
		s1[j]=L'\0';
		s2[j]=L'\0';
		res=wcscoll(s1,s2);
		if(res) return res;
		if(!c1 && c2)
			return -1;
		else if(!c2 && c1)
			return +1;
		else if(!c1 && !c2)
			++i;
	}

	if(bws1->len<bws2->len)
		return -1;
	else if(bws1->len>bws2->len)
		return +1;

	return 0;
}

/*
 * Write binary string to the file.
 * The output is ended either with '\n' (nl == true)
 * or '\0' (nl == false).
 */
int bwsfwrite(const bwstring* bws, FILE* f, bool nl)
{
	if(bws && f) {
		int printed=0;
		while(printed<(int)bwslen(bws)) {
			const wchar_t *s=bws_getcstr(bws)+printed;
			if(*s==L'\0') {
				int nums=fwprintf(f,L"%lc",*s);
				if(nums!=1)
					return -1;
				++printed;
			} else {
				int nums=fwprintf(f,L"%ls",s);
				if(nums<1)
					return -1;
				printed+=nums;
			}
		}
		if(nl) {
			fwprintf(f,L"%lc",L'\n');
		} else {
			fwprintf(f,L"%lc",L'\0');
		}
		return printed+1;
	}
	return -1;
}

/*
 * Allocate and read a binary string from file.
 * The strings are nl-ended or zero-ended, depending on the sort setting.
 */
bwstring *bwsfgetln(FILE* f, size_t *len, bool zero_ended)
{
	if (!zero_ended) {
		wchar_t *ret = fgetwln(f, len);
		if(!ret)
			return NULL;
		if(*len>0) {
			if(ret[*len-1]==L'\n') {
				ret[*len-1]=L'\0';
				--(*len);
			}
		}
		return bwssbdup(ret,*len);
	} else {
			static wchar_t* fgetwln_z_buffer = NULL;
			static size_t fgetwln_z_buffer_size = 0;

			*len = 0;

			if(feof(f)) return NULL;

			if (2 >= fgetwln_z_buffer_size) {
				fgetwln_z_buffer_size += 256;
				fgetwln_z_buffer = sort_realloc(fgetwln_z_buffer, sizeof(wchar_t) * fgetwln_z_buffer_size);
			}
			fgetwln_z_buffer[*len] = 0;

			while (!feof(f)) {
				wchar_t c = 0;
				if (*len + 2 >= fgetwln_z_buffer_size) {
					fgetwln_z_buffer_size += 256;
					fgetwln_z_buffer = sort_realloc(fgetwln_z_buffer, sizeof(wchar_t)
									* fgetwln_z_buffer_size);
				}
				c = fgetwc(f);
				if (c == WEOF) {
					if(*len==0) return NULL;
					break;
				} else if (c == L'\0') {
					break;
				}
				fgetwln_z_buffer[*len] = c;
				*len += 1;
				fgetwln_z_buffer[*len] = 0;
			}

			/* we do not count the last 0 */

			return bwssdup(fgetwln_z_buffer);
		}
}
