/*-
 * Copyright (C) 2009 Gabor Kovesdan <gabor@FreeBSD.org>
 * 2012 Oleg Moskalenko oleg.moskalenko@citrix.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/limits.h>

#include <errno.h>
#include <err.h>
#include <langinfo.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <wctype.h>

#include "sort.h"

#define	MAX(a, b)	(((a) > (b)) ? (a) : (b))

static wchar_t **months;

static int wstrcoll(key_value *kv1, key_value *kv2);
static int gnumcoll(key_value*, key_value *);
static int monthcoll(key_value*, key_value *);
static int numcoll(key_value*, key_value *);

/*
 * Allocate set of keys object
 */
keys_set* alloc_keys_set(void)
{
	static size_t setsz = 0;
	keys_set *key = (keys_set*) sort_malloc(sizeof(keys_set));
	if (!key)
		err(2, NULL);
	memset(key, 0, sizeof(keys_set));

	key->nums = keys_num;
	if (!setsz)
		setsz = keys_num * sizeof(key_value);

	key->set = (key_value*) sort_malloc(setsz);
	if (!(key->set))
		err(2, NULL);

	memset(key->set, 0, setsz);

	return key;
}

/*
 * De-allocate set of keys object
 */
void free_keys_set(keys_set* ks)
{
	if (ks) {
		if (ks->set) {
			size_t i = 0;
			for (i = 0; i < ks->nums; i++) {
				bwstring* k = ks->set[i].k;
				if (k) {
					bwsfree(k);
				}
			}
			sort_free(ks->set);
		}
		sort_free(ks);
	}
}

/*
 * Set value of a key in the keys set
 */
void set_key_on_keys_set(keys_set *ks, bwstring *s, size_t ind)
{
	if (ks && ks->set && ks->nums > ind) {
		key_value *kv = &(ks->set[ind]);
		if (kv->k) {
			bwsfree(kv->k);
		}
		kv->k = s;
	}
}

/*
 * Initialize a sort list item
 */
void sort_list_item_init(sort_list_item* si)
{
	if (si) {
		si->ks = NULL;
		si->str = NULL;
	}
}

/*
 * Calculate key for a sort list item
 */
static inline size_t sort_list_item_make_key(sort_list_item *si)
{
	if (!(si->ks)) {
		si->ks = preproc(si->str);
	}
	if(si->ks) {
		size_t i = 0;
		size_t mem=0;
		for (i = 0; i < si->ks->nums; i++) {
			bwstring* k = si->ks->set[i].k;
			if (k) {
				mem += bws_memsize(k) + sizeof(key_value);
			} else {
				mem += sizeof(key_value);
			}
		}
		return mem;
	}
	return 0;
}

/*
 * Set value of a sort list item.
 * Return combined string and keys memory size.
 */
size_t sort_list_item_set(sort_list_item* si, bwstring *str)
{
	if (si) {
		if (si->str) {
			if (si->str == str) {
				/* ??? */
				return 0;
			} else {
				bwsfree(si->str);
				si->str = NULL;
			}
		}
		if (si->ks) {
			free_keys_set(si->ks);
			si->ks = NULL;
		}
		si->str = str;
		return sizeof(sort_list_item) + bws_memsize(str) + sort_list_item_make_key(si);
	}
	return 0;
}

/*
 * De-allocate a sort list item object memory
 */
void sort_list_item_clean(sort_list_item* si)
{
	if (si) {
		if (si->str) {
			bwsfree(si->str);
			si->str = NULL;
		}
		if (si->ks) {
			free_keys_set(si->ks);
			si->ks = NULL;
		}
	}
}

/*
 * Swap two sort list item objects
 */
void sort_list_item_swap(sort_list_item* si1, sort_list_item* si2)
{
	if (si1 && si2) {
		bwstring *tmp;
		keys_set *kstmp;

		tmp = si1->str;
		si1->str = si2->str;
		si2->str = tmp;

		kstmp = si1->ks;
		si1->ks = si2->ks;
		si2->ks = kstmp;
	}
}

/*
 * Rips out leading blanks (-b).
 */
static bwstring *ignore_leading_blanks(bwstring *str)
{

	wchar_t *src = bws_getstr(str);
	wchar_t *dst = src;

	while (!bwsend(str,src) && iswblank(src[0]))
		++src;

	if (src != dst) {
		size_t newlen = bwslen(str)-(src-dst);
		while (!bwsend(str,src)) {
			*dst = *src;
			++dst;
			++src;
		}
		bws_setlen(str,newlen);

	}

	return str;
}

/*
 * Rips out nonprinting characters (-i).
 */
static bwstring *ignore_nonprinting(bwstring *str)
{

	wchar_t *src = bws_getstr(str);
	wchar_t *dst = src;
	wchar_t c = 0;
	size_t newlen=bwslen(str);

	while (!bwsend(str,src)) {
		c=*src;
		if (iswprint(c)) {
			*dst = c;
			++dst;
			++src;
		} else {
			++src;
			--newlen;
		}
	}

	bws_setlen(str,newlen);

	return str;
}

/*
 * Rips out any characters that are not alphanumeric characters
 * nor blanks (-d).
 */
static bwstring *dictionary_order(bwstring *str)
{

	wchar_t *src = bws_getstr(str);
	wchar_t *dst = src;
	wchar_t c = 0;
	size_t newlen=bwslen(str);

	while (!bwsend(str,src)) {
		c = *src;
		if (iswalnum(c) || iswblank(c)) {
			*dst = c;
			++dst;
			++src;
		} else {
			++src;
			--newlen;
		}
	}

	bws_setlen(str,newlen);

	return str;
}

/*
 * Converts string to lower case(-f).
 */
static bwstring *ignore_case(bwstring *str)
{
	wchar_t *s = bws_getstr(str);

	while (!bwsend(str,s)) {
		*s=towlower(*s);
		++s;
	}

	return str;
}

/*
 * Skip columns according to specs
 */
static size_t skip_cols_to_start(const bwstring *s, size_t cols, size_t start, bool skip_blanks, bool *empty_key)
{
	if(cols<1) {
		return bwslen(s)+1;
	}

	if(skip_blanks) {
		while(start<bwslen(s) && iswblank(bws_get(s,start))) ++start;
	}
	while(start<bwslen(s) && cols>1) {
		--cols; ++start;
	}
	if(start>=bwslen(s))
		*empty_key=true;

	return start;
}

/*
 * Skip fields according to specs
 */
static size_t skip_fields_to_start(const bwstring *s, size_t fields, bool *empty_field)
{
	if (fields < 2) {
		if (!bwslen(s))
			*empty_field = true;
		return 0;
	} else if (!(sort_opts_vals.tflag)) {
		size_t cpos = 0;
		bool pb = true;
		while(cpos<bwslen(s)) {
			bool isblank = iswblank(bws_get(s,cpos));
			if (isblank && !pb) {
				--fields;
				if (fields <= 1) {
					return cpos;
				}
			}
			pb = isblank;
			++cpos;
		}
		if (fields > 1)
			*empty_field = true;
		return cpos;
	} else {
		size_t cpos = 0;
		wchar_t c;
		while (cpos<bwslen(s)) {
			c = bws_get(s,cpos);
			if (c == sort_opts_vals.field_sep) {
				--fields;
				if (fields <= 1)
					return (cpos + 1);
			}
			++cpos;
		}
		if (fields > 1)
			*empty_field = true;
		return cpos;
	}
}

/*
 * Find fields start
 */
static void find_field_start(const bwstring* s, key_specs *ks, size_t *field_start, size_t *key_start,
				bool *empty_field, bool *empty_key)
{
	*field_start = skip_fields_to_start(s, ks->f1, empty_field);
	if (!*empty_field)
		*key_start = skip_cols_to_start(s, ks->c1, *field_start, ks->pos1b, empty_key);
	else
		*empty_key=true;
}

/*
 * Find end key position
 */
static size_t find_field_end(const bwstring* s, key_specs *ks)
{
	size_t pos_end = 0;
	size_t next_field_start = 0;
	bool empty_field = false;
	bool empty_key = false;
	size_t f2 = ks->f2;

	if (f2 == 0) {
		return bwslen(s) + 1;
	} else {
		if (ks->c2 == 0) {
			next_field_start = skip_fields_to_start(s, f2 + 1, &empty_field);
			if ((next_field_start > 0) && sort_opts_vals.tflag && (sort_opts_vals.field_sep == bws_get(s,
							next_field_start - 1)))
				--next_field_start;
		} else {
			next_field_start = skip_fields_to_start(s, f2, &empty_field);
		}
	}

	if (empty_field || (next_field_start >= bwslen(s)))
		return bwslen(s) + 1;

	if (ks->c2) {
		pos_end = skip_cols_to_start(s, ks->c2, next_field_start, ks->pos2b, &empty_key);
		if (pos_end < bwslen(s))
			++pos_end;
	} else {
		pos_end = next_field_start;
	}

	return pos_end;
}

/*
 * Cut a field according to the key specs
 */
static bwstring *cut_field(const bwstring* s, key_specs *ks)
{
	bwstring* ret=NULL;

	if (s && ks) {
		size_t field_start = 0;
		size_t key_start = 0;
		bool empty_field = false;
		bool empty_key = false;
		size_t key_end;
		size_t sz;

		find_field_start(s, ks, &field_start, &key_start, &empty_field, &empty_key);

		if (empty_key) {
			sz=0;
		} else {
			key_end = find_field_end(s, ks);
			if(key_end < key_start)
				sz=0;
			else
				sz = key_end - key_start;
		}

		ret = bwsalloc(sz+1);
		if (sz)
			bwsnocpy(ret, s, key_start, sz);
	} else {
		ret = bwssdup(L"");
	}

	return ret;
}

/*
 * Preprocesses a line applying the necessary transformations
 * specified by command line options and returns the preprocessed
 * string, which can be used to compare.
 */
keys_set *preproc(const bwstring *s)
{
	keys_set* kset = alloc_keys_set();

	bwstring *ret;

	if (sort_opts_vals.kflag) {

		size_t i = 0;
		for (i = 0; i < keys_num; i++) {

			key_specs *kspecs = &(keys[i]);

			current_sort_mods=&(kspecs->sm);

			bwstring *key = cut_field(s, kspecs);

			if (current_sort_mods->dflag)
				key = dictionary_order(key);
			else if (current_sort_mods->iflag)
				key = ignore_nonprinting(key);
			if (current_sort_mods->fflag || current_sort_mods->Mflag)
				key = ignore_case(key);

			set_key_on_keys_set(kset, key, i);
		}

	} else {
		ret = bwsdup(s);
		if (current_sort_mods->bflag)
			ret = ignore_leading_blanks(ret);
		if (current_sort_mods->dflag)
			ret = dictionary_order(ret);
		else if (current_sort_mods->iflag)
			ret = ignore_nonprinting(ret);
		if (current_sort_mods->fflag || current_sort_mods->Mflag)
			ret = ignore_case(ret);

		set_key_on_keys_set(kset, ret, 0);
	}

	return kset;
}

cmpcoll_t get_sort_func(sort_mods *sm)
{
	if (sm->nflag)
		return numcoll;
	else if (sm->gflag)
		return gnumcoll;
	else if (sm->Mflag)
		return monthcoll;
	else
		return wstrcoll;
}

/*
 * Compares the given strings.  Returns a positive number if
 * the first precedes the second, a negative number if the second is
 * the preceding one, and zero if they are equal.  This function calls
 * the underlying collate functions, which done the actual comparison.
 */
int key_coll(const keys_set *ps1, const keys_set *ps2)
{
	int res = 0;
	size_t i = 0;

	for (i = 0; i < ps1->nums; ++i) {

		current_sort_mods = &(keys[i].sm);

		if (current_sort_mods->rflag) {
			res=current_sort_mods->func(&(ps2->set[i]),&(ps1->set[i]));
		} else {
			res=current_sort_mods->func(&(ps1->set[i]),&(ps2->set[i]));
		}

		if (res)
			break;
	}

	current_sort_mods = default_sort_mods;

	return res;
}

/*
 * Compare two strings.
 * Plain symbol-by-symbol comparison.
 */
int str_coll(const bwstring *s1, const bwstring *s2)
{
	if (current_sort_mods->rflag) {
		const bwstring *tmp = s1;

		s1 = s2;
		s2 = tmp;
	}

	return (current_sort_mods->fflag ? bwscasecoll(s1, s2) : bwscoll(s1, s2));
}

/*
 * Compare a string and a sort list item, according to the sort specs.
 */
int str_list_coll(const bwstring *str1, sort_list_item **ss2)
{
	keys_set *ks1 = NULL;
	int ret = 0;

	current_sort_mods = default_sort_mods;

	ks1 = preproc(str1);

	sort_list_item_make_key(*ss2);

	if(debug_sort)
		printf("; s1=<%S>, s2=<%S>",bws_getcstr(str1),bws_getcstr((*ss2)->str));

	ret = key_coll(ks1, ((*ss2)->ks));

	if(debug_sort)
		printf("; cmp1=%d", ret);

	free_keys_set(ks1);

	if (!ret && !(sort_opts_vals.sflag) && sort_opts_vals.complex_sort) {
		ret = str_coll(str1, ((*ss2)->str));
		if(debug_sort)
			printf("; cmp2=%d",ret);
	}

	if(debug_sort)
		printf("\n");

	return ret;
}

/*
 * Compare two sort list items, according to the sort specs.
 */
int list_coll(sort_list_item **ss1, sort_list_item **ss2)
{
	int ret = 0;

	current_sort_mods = default_sort_mods;

	ret = key_coll(((*ss1)->ks), ((*ss2)->ks));

	if(debug_sort)
		printf("; s1=<%S>, s2=<%S>; cmp1=%d; cs=%d\n",bws_getcstr((*ss1)->str),bws_getcstr((*ss2)->str),ret,(int)sort_opts_vals.complex_sort);

	if(ret) return ret;

	if (!(sort_opts_vals.sflag) && sort_opts_vals.complex_sort) {
		ret = str_coll(((*ss1)->str), ((*ss2)->str));
		if(debug_sort)
			printf("; cmp2=%d\n",ret);
	}

	return ret;
}

/*
 * Check whether the thousands separator is in valid position.
 * If yes, then sets "previous position" equal to current.
 * If previous position is -1 (never occurred before) then always true.
 */
static inline int good_thousands_separator(int n, int *pc)
{
	if (*pc < 0) {
		*pc = n;
		return 1;
	}
	int diff = n - *pc;
	if (diff < 4)
		return 0;
	if (diff == 4) {
		*pc = n;
		return 1;
	}
	/* case when some separators were missed, for example 1,000000,000 */
	return good_thousands_separator(n - 3, pc);
}

/*
 * Maximum size of a number in the string (before or after decimal point)
 */
#define MAX_NUM_SIZE (65535)

/*
 * Read string s and parse the string into a fixed-decimal-point number.
 * sign equals -1 if the number is negative (explicit plus is not allowed, according to "info sort".
 * The number part before decimal point is in the smain, after the decimal point is in sfrac.
 * tail is the pointer to the remainder of the string.
 */
static int read_number(const bwstring *s0, int *sign, wchar_t *smain, int *main_len, wchar_t *sfrac, int *frac_len)
{
	const wchar_t* s=bws_getcstr(s0);
	int cpos = 0;
	int prev_thousands_separator = -1;
	int prev_len = 0;
	const wchar_t *prev_s = s;

	/* always end the fraction with zero, even if we have no fraction */
	sfrac[0] = 0;

	while (iswblank(*s))
		++s;

	if (*s == symbol_negative_sign) {
		*sign = -1;
		++s;
	}
	while (iswnumber(*s) && (*s == L'0'))
		++s;

	while (*s && *main_len < MAX_NUM_SIZE) {
		if (iswnumber(*s)) {
			smain[*main_len] = *s;
			++s;
			*main_len += 1;
			++cpos;
		} else if (*s == symbol_thousands_sep) {
			if (good_thousands_separator(cpos, &prev_thousands_separator)) {
				++s;
				++cpos;
				prev_len = *main_len;
				prev_s = s;
			} else {
				break;
			}
		} else {
			break;
		}
	}

	if (!good_thousands_separator(cpos, &prev_thousands_separator)) {
		*main_len = prev_len;
		s = prev_s;
	}

	smain[*main_len] = 0;

	if (*s == symbol_decimal_point) {
		++s;
		while (iswnumber(*s) && *frac_len < MAX_NUM_SIZE) {
			sfrac[*frac_len] = *s;
			++s;
			*frac_len += 1;
		}
		sfrac[*frac_len] = 0;
	}

	if((*main_len + *frac_len) == 0)
			*sign=0;

	return 0;
}

/*
 * Static big strings for the parsed numbers (performance optimization).
 */
static wchar_t smain1[MAX_NUM_SIZE+1];
static wchar_t smain2[MAX_NUM_SIZE+1];
static wchar_t sfrac1[MAX_NUM_SIZE+1];
static wchar_t sfrac2[MAX_NUM_SIZE+1];

/*
 * Implements string sort.
 */
static int wstrcoll(key_value *kv1, key_value *kv2)
{
	if (debug_sort)
		printf("; k1=<%S>(%zu), k2=<%S>(%zu)", bws_getcstr(kv1->k), bwslen(kv1->k), bws_getcstr(kv2->k), bwslen(
						kv2->k));

	return bwscoll(kv1->k, kv2->k);
}

/*
 * Implements numeric sort (-n).
 */
static int numcoll(key_value *kv1, key_value *kv2)
{
	const bwstring *s1 = kv1->k, *s2 = kv2->k;
	int sign1 = 0, sign2 = 0, main1 = 0, main2 = 0, frac1 = 0, frac2 = 0;
	int cmp_res = 0;
	bool key1_read = false, key2_read = false;

	if(debug_sort)
		printf("; %s: k1=<%S>, k2=<%S>",__FUNCTION__,bws_getcstr(s1),bws_getcstr(s2));

	if (s1 == s2)
		return 0;

	if (kv1->kh.status == HS_UNINITIALIZED) {
		/* read the number from the string */
		read_number(s1, &sign1, smain1, &main1, sfrac1, &frac1);
		key1_read = true;
		kv1->kh.v.nh.n1 = wcstoull(smain1, NULL, 10);
		if (kv1->kh.v.nh.n1 != ULLONG_MAX) {
			kv1->kh.status = HS_INITIALIZED;
		} else {
			kv1->kh.status = HS_ERROR;
		}
		if (sign1 < 0)
			kv1->kh.v.nh.neg1 = true;
		else
			kv1->kh.v.nh.neg1 = false;
	}

	if (kv2->kh.status == HS_UNINITIALIZED) {
		/* read the number from the string */
		read_number(s2, &sign2, smain2, &main2, sfrac2, &frac2);
		key2_read = true;
		kv2->kh.v.nh.n1 = wcstoull(smain2, NULL, 10);
		if (kv2->kh.v.nh.n1 != ULLONG_MAX) {
			kv2->kh.status = HS_INITIALIZED;
		} else {
			kv2->kh.status = HS_ERROR;
		}
		if (sign2 < 0)
			kv2->kh.v.nh.neg1 = true;
		else
			kv2->kh.v.nh.neg1 = false;
	}

	if (kv1->kh.status == HS_INITIALIZED && kv2->kh.status == HS_INITIALIZED) {

		bool neg1 = kv1->kh.v.nh.neg1;
		bool neg2 = kv2->kh.v.nh.neg1;

		if (neg1 && !neg2)
			return -1;
		if (neg2 && !neg1)
			return +1;
		unsigned long long n1 = kv1->kh.v.nh.n1;
		unsigned long long n2 = kv2->kh.v.nh.n1;
		if (n1 < n2) {
			if (neg1)
				return +1;
			return -1;
		} else if (n1 > n2) {
			if (neg1)
				return -1;
			return +1;
		}
	}

	/* read the numbers from the strings */
	if(!key1_read)
		read_number(s1, &sign1, smain1, &main1, sfrac1, &frac1);
	if(!key2_read)
		read_number(s2, &sign2, smain2, &main2, sfrac2, &frac2);

	/* if both numbers are empty assume that the strings are numerically equal */
	if (main1 < 1 && main2 < 1 && frac1 < 1 && frac2 < 1) {
		return 0;
	}

	/* we know the result if the signs are different */
	if (sign1 < 0 && sign2 >= 0)
		return -1;
	if (sign1 >= 0 && sign2 < 0)
		return +1;

	/* if the main part is of different size, we know the result
	 * (because the leading zeros are removed) */
	if (main1 < main2)
		cmp_res = -1;
	else if (main1 > main2)
		cmp_res = +1;
	/* if the sizes are equal then string compare gives the correct result */
	else
		cmp_res = wcscoll(smain1, smain2);

	/* check fraction */
	if (!cmp_res) {
		cmp_res = wcscoll(sfrac1, sfrac2);
	}

	if (!cmp_res) {
		return 0;
	}

	/* reverse result if the signs are negative */
	if (sign1 < 0 && sign2 < 0) {
		cmp_res = -cmp_res;
	}

	return cmp_res;
}

/*
 * Check for minus infinity
 */
static inline bool huge_minus(double d, int err1)
{
	if (err1 == ERANGE) {
		if (d == -HUGE_VAL || d == -HUGE_VALF || d == -HUGE_VALL) {
			return +1;
		}
	}
	return 0;
}

/*
 * Check for plus infinity
 */
static inline bool huge_plus(double d, int err1)
{
	if (err1 == ERANGE) {
		if (d == HUGE_VAL || d == HUGE_VALF || d == HUGE_VALL) {
			return +1;
		}
	}
	return 0;
}

/*
 * Check whether a function is a NAN
 */
static bool is_nan(double d) 
{
  return ((d==NAN)||(isnan(d)));
}

/*
 * Compare two NANs
 */
static int cmp_nans(double d1, double d2)
{
  if(d1<d2) 
    return -1;
  if(d2>d2)
    return +1;
  return 0;
}

/*
 * Implements general numeric sort (-g).
 */
static int gnumcoll(key_value *kv1, key_value *kv2)
{
	const wchar_t *s1 = bws_getcstr(kv1->k), *s2 = bws_getcstr(kv2->k);
	double d1=0, d2=0;
	wchar_t *ep1, *ep2;
	int err1=0, err2=0;
	bool key1_read = false, key2_read = false;

	if(debug_sort)
		printf("; %s: k1=<%S>, k2=<%S>",__FUNCTION__,s1,s2);

	if (kv1->kh.status == HS_UNINITIALIZED) {

		while (iswblank(*s1))
			++s1;

		errno = 0;
		d1 = wcstod(s1, &ep1);
		err1 = errno;

		if(s1==ep1) {
			kv1->kh.v.gh.notnum = true;
		} else if (err1 == 0) {
		  kv1->kh.v.gh.d = d1;
		  kv1->kh.v.gh.nan = is_nan(d1);
		  kv1->kh.status = HS_INITIALIZED;
		} else {
			kv1->kh.status = HS_ERROR;
		}

		key1_read = true;
	}

	if (kv2->kh.status == HS_UNINITIALIZED) {

		while (iswblank(*s2))
			++s2;

		errno = 0;
		d2 = wcstod(s2, &ep2);
		err2 = errno;

		if(s2==ep2) {
		  kv2->kh.v.gh.notnum = true;
		} else if (err2 == 0) {
		  kv2->kh.v.gh.d = d2;
		  kv2->kh.v.gh.nan = is_nan(d2);
		  kv2->kh.status = HS_INITIALIZED;
		} else {
		  kv2->kh.status = HS_ERROR;
		}

		key2_read = true;
	}

	if(kv1->kh.status==HS_INITIALIZED && kv2->kh.status==HS_INITIALIZED) {

		if(kv1->kh.v.gh.notnum) {
			if(kv2->kh.v.gh.notnum) {
				return 0;
			} else {
				return -1;
			}
		} else if(kv2->kh.v.gh.notnum) {
			return +1;
		}

		if(kv1->kh.v.gh.nan) {
			if(kv2->kh.v.gh.nan) {
			  return cmp_nans(kv1->kh.v.gh.d,kv2->kh.v.gh.d);;
			} else {
			  return -1;
			}
		} else if(kv2->kh.v.gh.nan) {
			return +1;
		}

		d1=kv1->kh.v.gh.d;
		d2=kv2->kh.v.gh.d;

		if(d1<d2)
			return -1;
		else if(d1>d2)
			return +1;
		else
			return 0;
	}

	if (!key1_read) {
		while (iswblank(*s1))
			++s1;

		errno = 0;
		d1 = wcstod(s1, &ep1);
		err1 = errno;
	}

	if (!key2_read) {
		while (iswblank(*s2))
			++s2;

		errno = 0;
		d2 = wcstod(s2, &ep2);
		err2 = errno;
	}

	/* Non-value case: */
	if (ep1 == s1) {
		if (ep2 == s2) {
			return 0;
		} else {
			return -1;
		}
	} else if (ep2 == s2) {
		return +1;
	}

	/* NAN case */
	if (is_nan(d1)) {
	  if (is_nan(d2)) {
	    return cmp_nans(d1,d2);
	  } else {
	    return -1;
	  }
	} else if (is_nan(d2)) {
	  return +1;
	}

	/* Infinities */
	if (err1 == ERANGE || err2 == ERANGE) {
		/* Minus infinity case */
		if (huge_minus(d1, err1)) {
			if (huge_minus(d2, err2)) {
				if (d1 < d2)
					return -1;
				if (d1 > d2)
					return +1;
				return 0;
			} else {
				return -1;
			}
		} else if (huge_minus(d2, err2)) {
			if (huge_minus(d1, err1)) {
				if (d1 < d2)
					return -1;
				if (d1 > d2)
					return +1;
				return 0;
			} else {
				return +1;
			}
		}

		/* Plus infinity case */
		if (huge_plus(d1, err1)) {
			if (huge_plus(d2, err2)) {
				if (d1 < d2)
					return -1;
				if (d1 > d2)
					return +1;
				return 0;
			} else {
				return +1;
			}
		} else if (huge_plus(d2, err2)) {
			if (huge_plus(d1, err1)) {
				if (d1 < d2)
					return -1;
				if (d1 > d2)
					return +1;
				return 0;
			} else {
				return -1;
			}
		}
	}

	if (d1 < d2)
		return -1;
	if (d1 > d2)
		return +1;

	return 0;
}

/*
 * A helper function for monthcoll.  If a line matches
 * a month name, it returns (number of the month - 1),
 * while if there is no match, it just return -1.
 */

static int month_score(const wchar_t *s)
{
	char *tmp;
	size_t len;

	while (iswblank(*s))
		++s;

	len = wcslen(s);

	if (months == NULL) {

		nl_item item[12] = { ABMON_1, ABMON_2, ABMON_3, ABMON_4, ABMON_5, ABMON_6, ABMON_7, ABMON_8, ABMON_9,
				ABMON_10, ABMON_11, ABMON_12 };

		wchar_t *m;
		int i = 0;

		months = sort_malloc(sizeof(wchar_t *) * 12);
		for (i = 0; i < 12; i++) {
			unsigned int j = 0;
			months[i]=NULL;
			tmp = nl_langinfo(item[i]);
			if(!tmp)
				continue;
			len = strlen(tmp);
			if(len<1)
				continue;
			while(iswblank(*tmp)) ++tmp;
			m = sort_malloc(SIZEOF_WCHAR_STRING(len + 1));
			if (mbstowcs(m, tmp, len) == ((size_t) -1))
				continue;
			m[len] = L'\0';
			for (j = 0; j < len; j++)
				m[j] = towlower(m[j]);
			months[i] = m;
		}
	}

	{
		int i = 0;
		for (i = 11; i >= 0; --i) {
			if (months[i] && (s == wcsstr(s, months[i])))
				return (i);
		}
	}

	return (-1);
}

/*
 * Implements month sort (-M).
 */
static int monthcoll(key_value* kv1, key_value *kv2)
{
	const wchar_t *s1 = bws_getcstr(kv1->k), *s2 = bws_getcstr(kv2->k);

	int val1 = 0;
	int val2 = 0;
	bool key1_read = false, key2_read = false;

	if(debug_sort)
		printf("; %s: k1=<%S>, k2=<%S>",__FUNCTION__,s1,s2);

	if (kv1->kh.status == HS_UNINITIALIZED) {
		kv1->kh.v.Mh.m = month_score(s1);
		key1_read = true;
		kv1->kh.status = HS_INITIALIZED;
	}

	if (kv2->kh.status == HS_UNINITIALIZED) {
		kv2->kh.v.Mh.m = month_score(s2);
		key2_read = true;
		kv2->kh.status = HS_INITIALIZED;
	}

	if (kv1->kh.status == HS_INITIALIZED) {
		val1 = kv1->kh.v.Mh.m;
		key1_read = true;
	}

	if (kv2->kh.status == HS_INITIALIZED) {
		val2 = kv2->kh.v.Mh.m;
		key2_read = true;
	}

	if(!key1_read)
		val1 = month_score(s1);
	if(!key2_read)
		val2 = month_score(s2);

	if (val1 == val2) {
		return 0;
	}
	if (val1 < val2)
		return -1;
	return +1;
}

