/*-
 * Copyright (C) 2009 Gabor Kovesdan <gabor@FreeBSD.org>
 * 2012 Oleg Moskalenko oleg.moskalenko@citrix.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/queue.h>

#include <err.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <wchar.h>

#include "sort.h"

/*
 * List elements of "cleanable" files list.
 */
struct CLEANABLE_FILE
{
	char* fn;
	LIST_ENTRY(CLEANABLE_FILE) files;
};

/*
 * List header of "cleanable" files list.
 */
static LIST_HEAD(CLEANABLE_FILES,CLEANABLE_FILE) tmp_files;

/*
 * Init tmp files list
 */
void init_tmp_files(void)
{
	LIST_INIT(&tmp_files);
}

/*
 * Save name of a tmp file for signal cleanup
 */
static void tmp_file_atexit(const char* tmp_file)
{
	if (tmp_file) {
		struct CLEANABLE_FILE *item = sort_malloc(sizeof(struct CLEANABLE_FILE));
		item->fn = strdup(tmp_file);
		LIST_INSERT_HEAD(&tmp_files, item, files);
	}
}

/*
 * Clear tmp files
 */
void clear_tmp_files(void)
{
	struct CLEANABLE_FILE *item = NULL;
	LIST_FOREACH(item,&tmp_files,files) {
		if ((item) && (item->fn))
			unlink(item->fn);
	}
}

/*
 * Read zero-terminated line from a file
 */
char* read_file0_line(FILE* f)
{
	static char* current_line = NULL;
	static size_t current_sz = 0;

	int c = 0;
	size_t pos = 0;

	if (!f || feof(f))
		return NULL;

	if (current_line && current_sz > 0)
		current_line[0] = 0;

	while (!feof(f)) {
		c = fgetc(f);
		if (feof(f) || (c == -1))
			break;
		if ((pos + 1) >= current_sz) {
			size_t newsz = (current_sz + 2) * 2;
			current_line = (char*) sort_realloc(current_line, newsz);
			if (!current_line)
				err(2, NULL);
			current_sz = newsz;
		}
		current_line[pos] = (char) c;
		if (!c)
			break;
		else
			current_line[pos + 1] = 0;
		++pos;
	}

	return current_line;
}

/*
 * Generate new temporary file name
 */
char* new_tmp_file_name(void)
{
	static unsigned int tfcounter = 0;
	const char* fn = ".bsdsort.";
	size_t sz = strlen(tmpdir) + 1 + strlen(fn) + 32 + 1;
	char* ret = (char*) sort_malloc(sz);
	sprintf(ret, "%s/%s%d.%u", tmpdir, fn, (int)getpid(), tfcounter++);
	tmp_file_atexit(ret);
	return ret;
}

/*
 * Initialize file list
 */
void file_list_init(file_list *fl, bool tmp)
{
	if (fl) {
		fl->count = 0;
		fl->sz = 0;
		fl->fns = NULL;
		fl->tmp = tmp;
	}
}

/*
 * Add a file name to the list
 */
void file_list_add(file_list* fl, char* fn, bool allocate)
{
	if (fl && fn) {
		if (fl->count >= fl->sz || (!(fl->fns))) {
			fl->sz = (fl->sz) * 2 + 1;
			fl->fns = (char**) sort_realloc(fl->fns, fl->sz * sizeof(char*));
		}
		if (allocate) {
			fl->fns[fl->count] = strdup(fn);
		} else {
			fl->fns[fl->count] = fn;
		}
		fl->count += 1;
	}
}

/*
 * Populate file list from array of file names
 */
void file_list_populate(file_list* fl, int argc, char** argv, bool allocate)
{
	if (fl && argv) {
		int i = 0;
		for (i = 0; i < argc; i++) {
			file_list_add(fl, argv[i], allocate);
		}
	}
}

/*
 * Clean file list data and delete the files,
 * if this is a list of temporary files
 */
void file_list_clean(file_list *fl)
{
	if (fl) {
		if (fl->fns) {
			int i = 0;
			for (i = 0; i < fl->count; i++) {
				if (fl->fns[i]) {
					if (fl->tmp) {
						unlink(fl->fns[i]);
					}
					sort_free(fl->fns[i]);
					fl->fns[i] = 0;
				}
			}
			sort_free(fl->fns);
			fl->fns = NULL;
		}
		fl->sz = 0;
		fl->count = 0;
		fl->tmp = false;
	}
}

/*
 * Init sort list
 */
void sort_list_init(sort_list *l)
{
	if (l) {
		l->count = 0;
		l->size = 0;
		l->memsize = sizeof(sort_list);
		l->list = NULL;
	}
}

/*
 * Add string to sort list
 */
void sort_list_add(sort_list *l, bwstring* str)
{
	if (l && str) {
		int indx = l->count;
		if (!(l->list) || (indx >= l->size)) {
			int newsize = (l->size + 1) * 2;
			l->list = (sort_list_item**) sort_realloc(l->list, sizeof(sort_list_item*) * newsize);
			if (!(l->list)) {
				err(2, NULL);
			}
			l->memsize += (newsize - l->size) * sizeof(sizeof(sort_list_item*));
			l->size = newsize;
		}
		l->list[indx] = (sort_list_item*) sort_malloc(sizeof(sort_list_item));
		if (!(l->list[indx])) {
			err(2, "memory allocation error");
		}
		sort_list_item_init(l->list[indx]);
		l->memsize += sort_list_item_set(l->list[indx], str);
		l->count += 1;
	}
}

/*
 * Clean sort list data
 */
void sort_list_clean(sort_list *l)
{
	if (l) {
		if (l->list) {
			int i = 0;
			for (i = 0; i < l->count; i++) {
				sort_list_item* item = l->list[i];
				if (item) {
					sort_list_item_clean(l->list[i]);
					sort_free(l->list[i]);
					l->list[i] = NULL;
				}
			}
			sort_free(l->list);
			l->list = NULL;
		}
		l->count = 0;
		l->size = 0;
		l->memsize = sizeof(sort_list);
	}
}

/*
 * Write sort list to file
 */
void sort_list_dump(sort_list *l, const char* fn)
{
	if (l && fn) {
		FILE* f = openfile(fn, "w");
		if (!f) {
			err(2, NULL);
		}

		if (l->list) {
			sort_list_item *last_printed = NULL;
			int i = 0;
			for (i = 0; i < l->count; i++) {
				sort_list_item* item = l->list[i];
				if (item) {
					if (item->str) {
						if (!(sort_opts_vals.uflag) || !last_printed || list_coll(
										&last_printed, &(l->list[i]))) {
							bwsfwrite(l->list[i]->str,f,!(sort_opts_vals.zflag));
							if (sort_opts_vals.uflag) {
								last_printed = l->list[i];
							}
						}
					}
				}
			}
		}

		fflush(f);
		if (f != stdout)
			fclose(f);
	}
}

/*
 * Checks if the given file is sorted.  Stops at the first disorder,
 * prints the disordered line and returns 1.
 */
int check(const char *fn)
{
	int res = 0;
	int pos = 1;

	bwstring *s1 = NULL, *s2 = NULL;
	keys_set *ks1 = NULL, *ks2 = NULL;

	FILE *file;
	size_t len;

	file = openfile(fn, "r");
	if (!file) {
		err(2, NULL);
		goto end;
	}

	s1 = bwsfgetln(file, &len, sort_opts_vals.zflag);
	if (!s1)
		goto end;

	ks1 = preproc(s1);

	s2 = bwsfgetln(file, &len, sort_opts_vals.zflag);
	if (!s2)
		goto end;

	ks2 = preproc(s2);

	for (; !feof(file);) {

		if(debug_sort)
			printf("s1=<%S>, s2=<%S>",bws_getcstr(s1),bws_getcstr(s2));
		int cmp = key_coll(ks2, ks1);
		if(debug_sort)
			printf("; cmp1=%d",cmp);

		if (!cmp && sort_opts_vals.complex_sort && !(sort_opts_vals.uflag) && !(sort_opts_vals.sflag)) {
			cmp = str_coll(s2, s1);
			if(debug_sort)
				printf("cmp2=%d",cmp);
		}
		if(debug_sort)
			printf("\n");

		if ((sort_opts_vals.uflag && (cmp <= 0)) || (cmp < 0)) {
			if (!(sort_opts_vals.csilentflag))
				warnx("%s:%d: disorder: %ls", fn, pos + 1, bws_getcstr(s2));
			res = 1;
			goto end;
		}

		pos++;

		bwsfree(s1);
		s1=s2;

		free_keys_set(ks1);
		ks1 = ks2;
		ks2 = NULL;

		s2 = bwsfgetln(file, &len, sort_opts_vals.zflag);
		if (!s2)
			goto end;

		ks2 = preproc(s2);
	}

	end:

	if (s1)
		bwsfree(s1);
	if (ks1)
		free_keys_set(ks1);
	if (s2)
		bwsfree(s2);
	if (ks2)
		free_keys_set(ks2);

	fclose(file);

	if (res)
		exit(res);

	return (0);
}

/*
 * Opens a file.  If the given filename is "-", stdout will be
 * opened.
 */
FILE *openfile(const char *fn, const char *mode)
{
	FILE *file;

	if (strcmp(fn, "-") == 0) {
		if (mode && mode[0] == 'r')
			return (stdin);
		else
			return (stdout);
	} else {
		if ((file = fopen(fn, mode)) == NULL)
			err(2, NULL);
	}

	return (file);
}

/*
 * Reads a file into the internal buffer.
 */
int procfile(FILE *file, sort_list *list, file_list *fl)
{

	bwstring *item;
	size_t len;

	while ((item=bwsfgetln(file, &len, sort_opts_vals.zflag)) != NULL) {

		sort_list_add(list, item);

		if (list->memsize >= available_free_memory) {
			char* fn = new_tmp_file_name();
			sort_list_to_file(list, fn);
			file_list_add(fl, fn, false);
			sort_list_clean(list);
		}
	}

	return (0);
}

/*
 * Structure to be used in file merge process.
 */
typedef struct _file_header
{
	FILE* f;
	sort_list_item *si; /* current top line */
	size_t file_pos;
} file_header;

/*
 * Compare file headers. Files with EOF are always pushed to the end of the list.
 */
static int file_header_cmp(file_header* f1, file_header* f2)
{
	if (f1 == f2)
		return 0;
	else {
		if (!(f1->f)) {
			if (!(f2->f))
				return 0;
			else
				return +1;
		} else if (!(f2->f))
			return -1;
		else {
			int ret = list_coll(&(f1->si), &(f2->si));
			if (!ret) {
				if (f1->file_pos < f2->file_pos)
					return -1;
				else
					return +1;
			}
			return ret;
		}
	}
}

/*
 * Allocate and init file header structure
 */
static void file_header_init(file_header** fh, const char* fn, size_t file_pos)
{
	if (fh && fn) {
		bwstring *line;
		size_t sz=0;
		*fh = (file_header*) sort_malloc(sizeof(file_header));
		if (!*fh) {
			err(2, NULL);
		}
		(*fh)->file_pos = file_pos;
		(*fh)->f = openfile(fn, "r");
		if (!((*fh)->f)) {
			perror(fn);
			err(2, "Cannot open file for reading");
		}
		line = bwsfgetln((*fh)->f, &sz, sort_opts_vals.zflag);
		if (!line) {
			fclose((*fh)->f);
			(*fh)->f = NULL;
			(*fh)->si = NULL;
		} else {
			(*fh)->si = (sort_list_item*)sort_malloc(sizeof(sort_list_item));
			if (!((*fh)->si)) {
				err(2, "memory allocation error");
			}
			sort_list_item_init((*fh)->si);
			sort_list_item_set((*fh)->si, line);
		}
	}
}

/*
 * Close file
 */
static void file_header_close(file_header** fh)
{
	if (fh && *fh) {
		if ((*fh)->f) {
			if ((*fh)->f != stdin)
				fclose((*fh)->f);
			(*fh)->f = NULL;
		}
		if ((*fh)->si) {
			sort_list_item_clean((*fh)->si);
			sort_free((*fh)->si);
			(*fh)->si = NULL;
		}
		sort_free(*fh);
		*fh = NULL;
	}
}

/*
 * Swap two array elements
 */
static void file_header_swap(file_header **fh, int i1, int i2)
{
	file_header* tmp = fh[i1];
	fh[i1] = fh[i2];
	fh[i2] = tmp;
}

/* heap algorithm ==>> */

/*
 * See heap sort algorithm
 * "Raises" last element to its right place
 */
static void file_header_heap_swim(file_header **fh, int indx)
{
	if (indx > 0) {
		int parent_index = (indx - 1) >> 1;
		if (file_header_cmp(fh[indx], fh[parent_index]) < 0) {
			/* swap child and parent and continue */
			file_header_swap(fh, indx, parent_index);
			file_header_heap_swim(fh, parent_index);
		}
	}
}

/*
 * Sink the top element to its correct position
 */
static void file_header_heap_sink(file_header **fh, int indx, int size)
{
	int left_child_index = indx + indx + 1;
	int right_child_index = left_child_index + 1;
	if (left_child_index < size) {
		int min_child_index = left_child_index;
		if ((right_child_index < size) && (file_header_cmp(fh[left_child_index], fh[right_child_index]) > 0)) {
			min_child_index = right_child_index;
		}
		if (file_header_cmp(fh[indx], fh[min_child_index]) > 0) {
			file_header_swap(fh, indx, min_child_index);
			file_header_heap_sink(fh, min_child_index, size);
		}
	}
}

/* <<== heap algorithm */

/* simple stable list insertion sort algorithm ==>> */

/*
 * Bubble sort: adds element to the sorted list, from right (bigger) end
 * We use it only once, so "bubble" sort is OK.
 */
static void file_header_list_swim(file_header **fh, int indx)
{
	if (indx > 0) {
		if (file_header_cmp(fh[indx - 1], fh[indx]) > 0) {
			file_header_swap(fh, indx - 1, indx);
			file_header_list_swim(fh, indx - 1);
		}
	}
}

/*
 * Find a new element position for efficient insertion sort (see next function).
 */
static int file_header_list_get_new_position(file_header **fh, int left, int right)
{
	if (right == left)
		return -1;
	else if (left + 1 == right)
		return left;
	else {
		int k = (right + left) >> 1;
		if (file_header_cmp(fh[0], fh[k]) >= 0) {
			return file_header_list_get_new_position(fh, k, right);
		} else {
			return file_header_list_get_new_position(fh, left, k);
		}
	}
}

/*
 * Insertion sort: adds elements to the sorted list, from the "left" (smaller) end.
 */
static void file_header_list_sink(file_header **fh, int indx, int size)
{
	int newpos = file_header_list_get_new_position(fh, indx, size);
	if (newpos > 0) {
		int cpos = 0;
		while (cpos < newpos) {
			file_header_swap(fh, cpos, cpos + 1);
			++cpos;
		}
	}
}

/* <<== simple stable list insertion sort algorithm */

/*
 * Adds element to the "left" end
 */
static void file_header_list_rearrange_from_header(file_header **fh, int size)
{
	if (sort_opts_vals.sflag) {
		file_header_list_sink(fh, 0, size);
	} else {
		file_header_heap_sink(fh, 0, size);
	}
}

/*
 * Adds element to the "right" end
 */
static void file_header_list_push(file_header* f, file_header **fh, int size)
{
	fh[size++] = f;
	if (sort_opts_vals.sflag) {
		file_header_list_swim(fh, size - 1);
	} else {
		file_header_heap_swim(fh, size - 1);
	}
}

/*
 * Prints the current line of the file
 */
static void file_header_print(file_header* fh, FILE* f_out)
{

	static bwstring *last_printed_str = NULL;

	if (fh && fh->f && f_out && fh->si && fh->si->str) {
		if (sort_opts_vals.uflag) {
			if (!last_printed_str || (str_list_coll(last_printed_str, &(fh->si)))) {
				bwsfwrite(fh->si->str,f_out,!(sort_opts_vals.zflag));
				if(last_printed_str) bwsfree(last_printed_str);
				last_printed_str = bwsdup(fh->si->str);
			}
		} else {
			bwsfwrite(fh->si->str,f_out,!(sort_opts_vals.zflag));
		}
	}
}

/*
 * Read next line
 */
static void file_header_read_next(file_header *fh)
{
	if (fh && fh->f) {
		size_t sz = 0;
		bwstring *tmp = bwsfgetln(fh->f, &sz, sort_opts_vals.zflag);
		if (!tmp) {
			fclose(fh->f);
			fh->f = NULL;
			if (fh->si) {
				sort_list_item_clean(fh->si);
				sort_free(fh->si);
				fh->si = NULL;
			}
		} else {
			if (!(fh->si)) {
				fh->si = (sort_list_item*) sort_malloc(sizeof(sort_list_item));
				sort_list_item_init(fh->si);
			}
			sort_list_item_set(fh->si,tmp);
		}
	}
}

/*
 * Merge array of "files headers"
 */
static void file_headers_merge(int fnum, file_header **fh, FILE* f_out)
{
	int i = 0;
	/* construct the initial heap structure (unstable merge) or list (stable merge): */
	for (i = 0; i < fnum; i++) {
		file_header_list_push(fh[i], fh, i);
	}
	while (fh[0]->f) { /* unfinished files are always in front */
		/* output the smallest line: */
		file_header_print(fh[0], f_out);
		/* read a new line, if possible: */
		file_header_read_next(fh[0]);
		/* re-arrange the list: */
		file_header_list_rearrange_from_header(fh, fnum);
	}
}

/*
 * Merges the given files into the output file, which can be
 * stdout.
 */
static void merge_files_array(int argc, char** argv, const char *fn_out)
{

	if (argv && fn_out) {

		FILE *f_out;
		int i;
		file_header **fh;

		f_out = openfile(fn_out, "w");

		if (!f_out) {
			err(2, NULL);
		}

		fh = (file_header**) sort_malloc((argc + 1) * sizeof(file_header*));

		for (i = 0; i < argc; i++) {
			file_header_init(fh + i, argv[i], (size_t) i);
		}

		file_headers_merge(argc, fh, f_out);

		for (i = 0; i < argc; i++) {
			file_header_close(fh + i);
		}

		sort_free(fh);

		fflush(f_out);
		if (f_out != stdout)
			fclose(f_out);

	}
}

/*
 * Shrinks the file list until its size smaller than max number of opened files
 */
static int shrink_file_list(file_list* fl)
{
	if (!fl || (size_t) (fl->count) < max_open_files)
		return 0;
	else {
		file_list new_fl;
		int indx = 0;
		file_list_init(&new_fl, true);
		while (indx < fl->count) {
			int num = fl->count - indx;
			char* fnew = new_tmp_file_name();
			if ((size_t) num >= max_open_files)
				num = max_open_files - 1;
			merge_files_array(num, fl->fns + indx, fnew);
			if (fl->tmp) {
				int i = 0;
				for (i = 0; i < num; i++) {
					unlink(fl->fns[indx + i]);
				}
			}
			file_list_add(&new_fl, fnew, false);
			indx += num;
		}
		fl->tmp = false; /* already taken care of */
		file_list_clean(fl);

		fl->count = new_fl.count;
		fl->fns = new_fl.fns;
		fl->sz = new_fl.sz;
		fl->tmp = new_fl.tmp;

		return 1;
	}
}

/*
 * Merge list of files
 */
void merge_files(file_list *fl, const char *fn_out)
{

	if (fl && fn_out) {

		while (shrink_file_list(fl))
			;

		merge_files_array(fl->count, fl->fns, fn_out);
	}
}

static const char* get_sort_method_name(int sm)
{
	if (sm == SORT_MERGESORT) {
		return "mergesort";
	} else if (sort_opts_vals.sort_method == SORT_RADIXSORT) {
		return "radixsort";
	} else if (sort_opts_vals.sort_method == SORT_HEAPSORT) {
		return "heapsort";
	} else {
		return "quicksort";
	}
}

/*
 * Sort list of lines and writes it to the file
 */
void sort_list_to_file(sort_list *list, const char* outfile)
{
	if (keys_num == 1 && !(current_sort_mods->Mflag) && !(current_sort_mods->Rflag) && !(current_sort_mods->Vflag)
					&& !(current_sort_mods->gflag) && !(current_sort_mods->hflag)
					&& !(current_sort_mods->nflag) && (MB_CUR_MAX == 1)) {
		if (sort_opts_vals.sort_method == SORT_DEFAULT) {
			sort_opts_vals.sort_method = SORT_RADIXSORT;
		}
	} else if (sort_opts_vals.sort_method == SORT_RADIXSORT) {
		fprintf(stderr,"Radix sort cannot be used with these sort options and/or this locale setting\n");
		exit(2);
	}

	/* to handle stable sort and the unique cases in the right order, we need stable basic algorithm */
	if (sort_opts_vals.sflag) {
		if (sort_opts_vals.sort_method == SORT_MERGESORT) {
			;
		} else if (sort_opts_vals.sort_method == SORT_RADIXSORT) {
			;
		} else if (sort_opts_vals.sort_method == SORT_DEFAULT) {
			sort_opts_vals.sort_method = SORT_MERGESORT;
		} else {
			fprintf(stderr, "The chosen sort method cannot be used with stable and/or unique sort\n");
			exit(2);
		}
	}

	if (sort_opts_vals.sort_method == SORT_DEFAULT) {
		sort_opts_vals.sort_method = SORT_QSORT;
	}

	if(debug_sort)
		printf("sort_method=%s\n", get_sort_method_name(sort_opts_vals.sort_method));

	switch (sort_opts_vals.sort_method){
	case SORT_RADIXSORT:
		rxsort(list->list, list->count);
		break;
	case SORT_MERGESORT:
		if (mergesort(list->list, list->count, sizeof(sort_list_item *),
						(int(*)(const void *, const void *)) list_coll) == -1)
			err(2, NULL);
		break;
	case SORT_HEAPSORT:
		if (heapsort(list->list, list->count, sizeof(sort_list_item *),
						(int(*)(const void *, const void *)) list_coll) == -1)
			err(2, NULL);
		break;
	default:
		qsort(list->list, list->count, sizeof(sort_list_item *), (int(*)(const void *, const void *)) list_coll);
		break;
	}

	sort_list_dump(list, outfile);
}
