/* Copyright (C) 2012 Oleg Moskalenko oleg.moskalenko@citrix.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <sys/limits.h>

#include <errno.h>
#include <err.h>
#include <langinfo.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
#include <wchar.h>
#include <wctype.h>

#include "sort.h"

static wchar_t *ctable=NULL;
static unsigned char *indtable=NULL;

typedef struct _sort_level {

	size_t level;

	sort_list_item** tosort;
	size_t tosort_num;

	sort_list_item** leaves;
	size_t leaves_num;

	struct _sort_level** sublevels;
	size_t sln;
	size_t real_sln;
} sort_level;

static int cmpwchar(const void *pc1, const void *pc2)
{
	int res=0;
	const wchar_t *c1=(const wchar_t*)pc1;
	const wchar_t *c2=(const wchar_t*)pc2;
	wchar_t s1[2];
	wchar_t s2[2];
	s1[0]=*c1;
	s1[1]=L'\0';
	s2[0]=*c2;
	s2[1]=L'\0';

	res = wcscoll(s1,s2);

	return res;
}

static void make_char_table(void)
{
	if(!ctable) {
		int i=0;
		ctable=(wchar_t*)sort_malloc(SIZEOF_WCHAR_STRING(256));
		for(i=0;i<256;++i) {
			ctable[i]=i;
		}
		qsort(ctable,256,sizeof(wchar_t),cmpwchar);

		indtable=(unsigned char*)sort_malloc(256);
		memset(indtable,0,256);
		for(i=0;i<256;++i) {
			unsigned char c = (unsigned char)ctable[i];
			unsigned char pos = (unsigned char)i;
			indtable[c] = pos;
			/* printf("%s: c=%d, pos=%d\n",__FUNCTION__,(int)c,(int)pos); */
		}
	}
}

static void add_to_sublevel(sort_level* sl, sort_list_item *item, wchar_t wc)
{
	size_t indx = (unsigned char)wc;
	sort_level* ssl = sl->sublevels[indx];
	if(!ssl) {
		ssl=(sort_level*)sort_malloc(sizeof(sort_level));
		if(!ssl)
			err(2,NULL);
		memset(ssl,0,sizeof(sort_level));

		ssl->level=sl->level+1;
		sl->sublevels[indx] = ssl;

		++(sl->real_sln);
	}

	ssl->tosort=(sort_list_item**)sort_realloc(ssl->tosort,sizeof(sort_list_item*)*(++(ssl->tosort_num)));
	ssl->tosort[ssl->tosort_num-1]=item;
}

static inline void add_leaf(sort_level* sl, sort_list_item *item) {
	sl->leaves=(sort_list_item**)sort_realloc(sl->leaves,(sizeof(sort_list_item*)*(++(sl->leaves_num))));
	sl->leaves[sl->leaves_num-1]=item;
}

static inline int get_wc_index(sort_list_item* sli, size_t level) {
	if(sli->ks && sli->ks->nums==1) {
		bwstring *bws=sli->ks->set[0].k;
		if(bws && (bwslen(bws)>level)) {
			unsigned char c = (unsigned char)bws_get(bws,level);
			return indtable[c];
		}
	}
	return -1;
}

static inline void place_item(sort_level* sl, size_t item)
{
	sort_list_item* sli=sl->tosort[item];
	int c = get_wc_index(sli,sl->level);
	if(c == -1) {
		add_leaf(sl,sli);
	} else {
		add_to_sublevel(sl,sli,c);
	}
}

static const size_t slsz=256*sizeof(sort_level*);
static sort_level* tmp_sublevels[256];

static void run_sort_level_next(sort_level* sl)
{
	size_t i = 0;

	switch (sl->tosort_num){
	case 0:
		return;
	case (1):
		sl->leaves = sl->tosort;
		sl->leaves_num = sl->tosort_num;
		sl->tosort = NULL;
		sl->tosort_num = 0;
		return;
	case (2):
		sl->leaves = sl->tosort;
		sl->leaves_num = sl->tosort_num;
		sl->tosort = NULL;
		sl->tosort_num = 0;
		if (list_coll(&(sl->leaves[0]), &(sl->leaves[1])) > 0) {
			sort_list_item* tmp = sl->leaves[1];
			sl->leaves[1] = sl->leaves[0];
			sl->leaves[0] = tmp;
		}
		return;
	default:
		if ((sl->tosort_num < 33)||(sl->level>15)) {
			sl->leaves = sl->tosort;
			sl->leaves_num = sl->tosort_num;
			sl->tosort = NULL;
			sl->tosort_num = 0;
			if (mergesort(sl->leaves, sl->leaves_num, sizeof(sort_list_item *), (int(*)(const void *,
							const void *)) list_coll) == -1)
				err(2, "Radix sort error 3");
		}
	}

	sl->sln = 256;
	sl->sublevels = tmp_sublevels;
	memset(sl->sublevels, 0, slsz);

	sl->real_sln = 0;

	for (i = 0; i < sl->tosort_num; ++i) {
		place_item(sl, i);
	}

	sort_free(sl->tosort);
	sl->tosort = NULL;
	sl->tosort_num = 0;

	{
		size_t new_sln = sl->real_sln;
		size_t newsz = 0;
		size_t counter = 0;
		sort_level** new_sls = NULL;

		newsz = new_sln * sizeof(sort_level*);
		new_sls = (sort_level**) sort_malloc(newsz);
		for (i = 0; i < sl->sln; ++i) {
			if (sl->sublevels[i]) {
				new_sls[counter++] = sl->sublevels[i];
			}
		}

		sl->sublevels = new_sls;
		sl->sln = new_sln;
	}

	if (sl->leaves_num > 1) {
		if (mergesort(sl->leaves, sl->leaves_num, sizeof(sort_list_item *),
						(int(*)(const void *, const void *)) list_coll) == -1)
			err(2, "Radix sort error 2");
	}

	for (i = 0; i < sl->sln; ++i) {
		sort_level* slc=sl->sublevels[i];
		if(slc)
			run_sort_level_next(slc);
	}
}

static void run_sort_level(sort_level* sl)
{
	size_t i = 0;

	sl->sln = 256;
	sl->sublevels = tmp_sublevels;
	memset(sl->sublevels, 0, slsz);

	sl->real_sln = 0;

	for (i = 0; i < sl->tosort_num; ++i) {
		place_item(sl, i);
	}

	{
		size_t new_sln = sl->real_sln;
		size_t newsz = 0;
		size_t counter = 0;
		sort_level** new_sls = NULL;

		newsz = new_sln * sizeof(sort_level*);
		new_sls = (sort_level**) sort_malloc(newsz);
		for (i = 0; i < sl->sln; ++i) {
			if (sl->sublevels[i]) {
				new_sls[counter++] = sl->sublevels[i];
			}
		}

		sl->sublevels = new_sls;
		sl->sln = new_sln;
	}

	if (sl->leaves_num > 1) {
		if (mergesort(sl->leaves, sl->leaves_num, sizeof(sort_list_item *),
						(int(*)(const void *, const void *)) list_coll) == -1)
			err(2, "Radix sort error 2");
	}

	for (i = 0; i < sl->sln; ++i) {
		sort_level* slc=sl->sublevels[i];
		if(slc)
			run_sort_level_next(slc);
	}
}

static void walk_sort_level(sort_level *sl, sort_list_item** base, size_t *indx)
{
	if (sl) {
		size_t i = 0;
		if (current_sort_mods->rflag) {
			for (i = 0; i < sl->sln; ++i) {
				walk_sort_level(sl->sublevels[sl->sln - i - 1], base, indx);
			}
			for (i = 0; i < sl->leaves_num; ++i) {
				base[(*indx)++] = sl->leaves[i];
			}
		} else {
			for (i = 0; i < sl->leaves_num; ++i) {
				base[(*indx)++] = sl->leaves[i];
			}
			for (i = 0; i < sl->sln; ++i) {
				walk_sort_level(sl->sublevels[i], base, indx);
			}
		}
	}
}

static void free_sort_level(sort_level *sl)
{
	if(sl) {
		size_t i=0;
		if(sl->leaves) sort_free(sl->leaves);
		if(sl->level>0 && sl->tosort) sort_free(sl->tosort);
		for(i=0;i<sl->sln;++i)
			free_sort_level(sl->sublevels[i]);
		sort_free(sl);
	}
}

static void run_sort(sort_list_item **base, size_t nmemb)
{
	size_t n=0;
	sort_level *sl=(sort_level*)sort_malloc(sizeof(sort_level));
	if(!sl)
		err(2,NULL);
	memset(sl,0,sizeof(sort_level));

	sl->tosort=base;
	sl->tosort_num=nmemb;

	run_sort_level(sl);

	walk_sort_level(sl, sl->tosort, &n);

	if(n!=nmemb) {
		err(2,"Radix sort error 1");
	}

	free_sort_level(sl);
}

void rxsort(sort_list_item **base, size_t nmemb)
{
	make_char_table();
	run_sort(base,nmemb);
}
