/*-
 * Copyright (C) 2009 Gabor Kovesdan <gabor@FreeBSD.org>
 * 2012 Oleg Moskalenko oleg.moskalenko@citrix.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

//TODO: random-sort
//TODO: human-numeric-sort
//TODO: version-sort
//TODO: compress-program
//TODO: performance
//TODO: multi threaded sort

#include <sys/stat.h>
#include <sys/types.h>
#include <sys/sysctl.h>
#include <sys/stat.h>

#include <regex.h>
#include <signal.h>
#include <err.h>
#include <errno.h>
#include <getopt.h>
#include <limits.h>
#include <locale.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <wchar.h>
#include <wctype.h>

#include "sort.h"

#ifndef WITHOUT_NLS
#include <nl_types.h>
nl_catd catalog;
#endif

bool debug_sort=false;

void fix_obsolete_keys(int *argc, char** argv);

/*
 * Default messages to use when NLS is disabled or no catalogue
 * is found.
 */
const char *nlsstr[] = { "",
/* 1*/"you cannot use -%c and -%c together",
/* 2*/"extra argument not allowed with -c" };

extern char *__progname;

sort_opts sort_opts_vals;

static sort_mods default_sort_mods_object;
sort_mods* current_sort_mods = &default_sort_mods_object;
sort_mods* const default_sort_mods = &default_sort_mods_object;

key_specs* keys = NULL;
size_t keys_num = 0;

wchar_t symbol_decimal_point = L'.';
wchar_t symbol_thousands_sep = L',';
wchar_t symbol_negative_sign = L'-';
wchar_t symbol_positive_sign = L'+';

unsigned long long free_memory = 1000000;
unsigned long long available_free_memory = 250000;

const char* tmpdir = "/var/tmp";

size_t max_open_files = 16;

/*
 * Arguments from file (when file0-from option is used:
 */
static int argc_from_file0 = -1;
static char** argv_from_file0 = NULL;

/*
 * Placeholder symbols for options which have no single-character equivalent
 */
enum
{
	SORT_OPT = CHAR_MAX + 1, HELP_OPT, CP_OPT, FF_OPT, BS_OPT, VERSION_OPT, DEBUG_OPT,
	QSORT_OPT, MERGESORT_OPT, HEAPSORT_OPT, RADIXSORT_OPT
};

#define	OPTIONS	"cCRbdfgS:hik:T:Mmno:rst:uVz"

#define NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS (6)
static const char mutually_exclusive_flags[NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS] = { 'M', 'n', 'g', 'R', 'h', 'V' };

struct option long_options[] = {
				{ "ignore-leading-blanks", no_argument, NULL, 'b' },
				{ "check", optional_argument, NULL, 'c' },
				{ "check=silent|quiet", optional_argument, NULL, 'C' },
				{ "dictionary-order", no_argument, NULL, 'd' },
				{ "ignore-case", no_argument, NULL, 'f' },
				{ "general-numeric-sort", no_argument, NULL, 'g' },
				{ "ignore-nonprinting", no_argument, NULL, 'i' },
				{ "key", required_argument, NULL, 'k' },
				{ "month-sort", no_argument, NULL, 'M' },
				{ "numeric-sort", no_argument, NULL, 'n' },
				{ "random-sort", no_argument, NULL, 'R' }, /* unimplemented */
				{ "human-numeric-sort", no_argument, NULL, 'h' }, /* unimplemented */
				{ "merge", no_argument, NULL, 'm' },
				{ "output", required_argument, NULL, 'o' },
				{ "reverse", no_argument, NULL, 'r' },
				{ "stable", no_argument, NULL, 's' },
				{ "field-separator", required_argument, NULL, 't' },
				{ "buffer-size", required_argument, NULL, 'S' },
				{ "temporary-directory",required_argument, NULL, 'T' },
				{ "unique", no_argument, NULL, 'u' },
				{ "version-sort",no_argument, NULL, 'V' }, /* unimplemented */
				{ "zero-terminated", no_argument, NULL, 'z' },
				{ "sort", required_argument, NULL, SORT_OPT },
				{ "help",no_argument, NULL, HELP_OPT },
				{ "compress-program", required_argument, NULL, CP_OPT }, /* unimplemented  */
				{ "files0-from", required_argument, NULL, FF_OPT },
				{ "batch-size", required_argument, NULL, BS_OPT },
				{ "version", no_argument, NULL, VERSION_OPT },
				{ "debug", no_argument, NULL, DEBUG_OPT },
				{ "qsort", no_argument, NULL, QSORT_OPT },
				{ "mergesort", no_argument, NULL, MERGESORT_OPT },
				{ "heapsort", no_argument, NULL, HEAPSORT_OPT },
				{ "radixsort", no_argument, NULL, RADIXSORT_OPT },
				{ NULL, no_argument, NULL, 0 }
};

/*
 * Check where sort modifier is present
 */
static bool sort_modifier_empty(sort_mods *sm)
{
	if (!sm)
		return true;
	return !(sm->Mflag || sm->nflag || sm->gflag || sm->rflag ||
					sm->hflag || sm->dflag || sm->fflag);
}

/*
 * Print out usage text.
 */
static void usage(bool opt_err)
{
	FILE* out = stdout;
	struct option *o = &(long_options[0]);
	if (opt_err)
		out = stderr;
	fprintf(out, "Usage: %s [-cCRbdfnghiMmruVsztST] [-kPOS1[,POS2]] [+POS1 [-POS2]] [-o outfile] [file ...]\n", __progname);
	fprintf(out, "Long options:\n");
	while (o->name) {
		fprintf(out, "    --%s\n", o->name);
		++o;
	}
	fprintf(out,"For more information, see 'info sort' and IEEE Std 1003.1 page:\n http://pubs.opengroup.org/onlinepubs/009695399/utilities/sort.html\n");
	if (opt_err)
		exit(2);
	exit(0);
}

/*
 * Read input file names from a file (file0-from option).
 */
static void read_fns_from_file0(const char* fn)
{
	if (fn) {
		FILE* f = fopen(fn, "r");
		if (!f)
			err(2, NULL);
		while (!feof(f)) {
			char* line = read_file0_line(f);
			if (line && *line) {
				++argc_from_file0;
				if (argc_from_file0 < 1)
					argc_from_file0 = 1;
				argv_from_file0 = (char**) sort_realloc(argv_from_file0, argc_from_file0
								* sizeof(char*));
				if (!argv_from_file0)
					err(2, NULL);
				argv_from_file0[argc_from_file0 - 1] = strdup(line);
			}
		}
	}
}

/*
 * Check how much RAM is available for the sort.
 */
static void set_free_memory(void)
{
	size_t pages = 0;
	size_t psize = 0;
	size_t psz = sizeof(pages);
	size_t pszsz = sizeof(psize);
	if (sysctlbyname("vm.stats.vm.v_free_count", &pages, &psz, NULL, 0) < 0) {
		perror("vm.stats.vm.v_free_count");
		return;
	}
	if (sysctlbyname("vm.stats.vm.v_page_size", &psize, &pszsz, NULL, 0) < 0) {
		perror("vm.stats.vm.v_page_size");
		return;
	}
	free_memory = (unsigned long long) pages * (unsigned long long) psize;
	available_free_memory = free_memory >> 1;
}

/*
 * Convert "plain" symbol to wide symbol, with default value.
 */
static void conv_mbtowc(wchar_t *wc, const char* c, const wchar_t def)
{
	if (wc && c) {
		int res = mbtowc(wc, c, MB_CUR_MAX);
		if (res < 1)
			*wc = def;
	}
}

/*
 * Set current locale symbols.
 */
static void set_locale(void)
{

	struct lconv *lc;

	setlocale(LC_ALL, "");

	lc = localeconv();

	if (lc) {
		/* obtain LC_NUMERIC info */
		/* Convert to wide char form */
		conv_mbtowc(&symbol_decimal_point, lc->decimal_point, symbol_decimal_point);
		conv_mbtowc(&symbol_thousands_sep, lc->thousands_sep, symbol_thousands_sep);
		conv_mbtowc(&symbol_positive_sign, lc->positive_sign, symbol_positive_sign);
		conv_mbtowc(&symbol_negative_sign, lc->negative_sign, symbol_negative_sign);
	}
}

/*
 * Set directory temporary files.
 */
static void set_tmpdir(void)
{
	char* td = getenv("TMPDIR");
	if (td)
		tmpdir = strdup(td);
}

/*
 * Parse -S option.
 */
static unsigned long long parse_memory_buffer_value(const char* value)
{
	if (!value)
		return available_free_memory;
	else {
		char *endptr = NULL;
		unsigned long long membuf = strtoll(value, &endptr, 10);
		if (membuf == 0 || membuf == (unsigned long long) LLONG_MIN || membuf == (unsigned long long) LLONG_MAX)
			membuf = available_free_memory;
		else {
			switch (*endptr){
			case 'Y':
				membuf *= 1024;
			case 'Z':
				membuf *= 1024;
			case 'E':
				membuf *= 1024;
			case 'P':
				membuf *= 1024;
			case 'T':
				membuf *= 1024;
			case 'G':
				membuf *= 1024;
			case 'M':
				membuf *= 1024;
			case '\0':
			case 'K':
				membuf *= 1024;
			case 'b':
				break;
			case '%':
				membuf = (available_free_memory * membuf) / 100;
				break;
			default:
				fprintf(stderr, "%s: %s\n", strerror(EINVAL), optarg);
				membuf = available_free_memory;
			}
		}
		return membuf;
	}
}

/*
 * Signal handler that clears the temporary files.
 */
static void sig_handler(int sig, siginfo_t *siginfo, void *context)
{
	sig = sig;
	siginfo = siginfo;
	context = context;
	clear_tmp_files();
	exit(-1);
}

/*
 * Set signal handler on panic signals.
 */
static void set_signal_handler(void)
{

	struct sigaction sa;
	memset(&sa, 0, sizeof(sa));

	sa.sa_sigaction = &sig_handler;
	sa.sa_flags = SA_SIGINFO;

	if (sigaction(SIGTERM, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
	if (sigaction(SIGHUP, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
	if (sigaction(SIGINT, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
	if (sigaction(SIGQUIT, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
	if (sigaction(SIGABRT, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
	if (sigaction(SIGBUS, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
	if (sigaction(SIGSEGV, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
	if (sigaction(SIGUSR1, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
	if (sigaction(SIGUSR2, &sa, NULL) < 0) {
		perror("sigaction");
		return;
	}
}

/*
 * Print "unimplemented" message and exit with status 2.
 */
static void unimplemented(const char* what)
{
	fprintf(stderr, "Feature unimplemented: %s\n", what);
	exit(2);
}

/*
 * Print "unknown" message and exit with status 2.
 */
static void unknown(const char* what)
{
	fprintf(stderr, "Feature unknown: %s\n", what);
	exit(2);
}

/*
 * Check whether a contradictory input options are used.
 */
static void check_mutually_exclusive_flags(char c, bool *mef_flags)
{
	int i = 0;
	int mec = 0;
	bool found_others = false;
	int fo_index = 0;
	bool found_this = false;

	for (i = 0; i < NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS; i++) {

		mec = mutually_exclusive_flags[i];

		if (mec != c) {
			if (mef_flags[i]) {
				if (found_this)
					errx(1, getstr(1), c, mec);
				found_others = true;
				fo_index = i;
			}
		} else {
			if (found_others)
				errx(1, getstr(1), c, mutually_exclusive_flags[fo_index]);
			mef_flags[i] = true;
			found_this = true;
		}
	}
}

/*
 * Initialise sort opts data.
 */
static void set_sort_opts(void)
{
	memset(&default_sort_mods_object, 0, sizeof(default_sort_mods_object));
	memset(&sort_opts_vals, 0, sizeof(sort_opts_vals));
	default_sort_mods_object.func=get_sort_func(&default_sort_mods_object);
}

/*
 * Set a sort modifier on a sort modifiers object.
 */
static bool set_sort_modifier(sort_mods* sm, int c)
{
	if (sm) {
		switch (c){
		case 'b':
			sm->bflag = true;
			break;
		case 'd':
			sm->dflag = true;
			break;
		case 'f':
			sm->fflag = true;
			break;
		case 'g':
			sm->gflag = true;
			break;
		case 'i':
			sm->iflag = true;
			break;
		case 'R':
			sm->Rflag = true;
			unimplemented("random-sort");
			break;
		case 'M':
			sm->Mflag = true;
			break;
		case 'n':
			sm->nflag = true;
			break;
		case 'r':
			sm->rflag = true;
			break;
		case 'V':
			sm->Vflag = true;
			unimplemented("version-sort");
			break;
		case 'h':
			sm->hflag = true;
			unimplemented("human-numeric-sort");
			break;
		default:
			return false;
		}
		sort_opts_vals.complex_sort=true;
		sm->func=get_sort_func(sm);
	}
	return true;
}

/*
 * Parse POS in -k option.
 */
static int parse_pos(const char* s, key_specs* ks, bool *mef_flags, bool second)
{
	int ret=-1;
	regex_t re;
	size_t nmatch = 4;
	regmatch_t pmatch[4];
	char* f=NULL, *c=NULL;
	size_t len;

	const char* sregexp = "^([0-9]+)(\\.[0-9]+)?([bdfirMngRhV]+)?$";

	if (regcomp(&re, sregexp, REG_EXTENDED) != 0) {
		return -1;
	}

	if (regexec(&re, s, nmatch, pmatch, 0) != 0) {
		goto end;
	}

	if (pmatch[0].rm_eo <= pmatch[0].rm_so) {
		goto end;
	}

	if (pmatch[1].rm_eo <= pmatch[1].rm_so) {
		goto end;
	}

	len = pmatch[1].rm_eo - pmatch[1].rm_so;
	f = (char*) sort_malloc((len + 1) * sizeof(char));
	if (!f)
		err(2, NULL);

	strncpy(f, s + pmatch[1].rm_so, len);
	f[len] = 0;

	if (second) {
		ks->f2 = (size_t) strtoul(f, NULL, 10);
		if(ks->f2 == 0) {
			fprintf(stderr,"0 field in key specs\n");
			goto end;
		}
	} else {
		ks->f1 = (size_t) strtoul(f, NULL, 10);
		if (ks->f1 == 0) {
			fprintf(stderr,"0 field in key specs\n");
			goto end;
		}
	}

	if (pmatch[2].rm_eo > pmatch[2].rm_so) {
		len = pmatch[2].rm_eo - pmatch[2].rm_so - 1;
		c = (char*) sort_malloc((len + 1) * sizeof(char));
		if (!c)
			err(2, NULL);

		strncpy(c, s + pmatch[2].rm_so + 1, len);
		c[len] = 0;

		if (second)
			ks->c2 = (size_t) strtoul(c, NULL, 10);
		else {
			ks->c1 = (size_t) strtoul(c, NULL, 10);
			if (ks->c1 == 0) {
				fprintf(stderr,"0 column in key specs\n");
				goto end;
			}
		}
	} else {
		if (second)
			ks->c2 = 0;
		else
			ks->c1 = 1;
	}

	if (pmatch[3].rm_eo > pmatch[3].rm_so) {

		regoff_t i = 0;

		for (i = pmatch[3].rm_so; i < pmatch[3].rm_eo; i++) {
			check_mutually_exclusive_flags(s[i], mef_flags);
			if (s[i] == 'b') {
				if (second)
					ks->pos2b = true;
				else
					ks->pos1b = true;
			} else if (!set_sort_modifier(&(ks->sm), s[i])) {
				goto end;
			}
		}
	}

	ret=0;

	end:

	if(c) sort_free(c);
	if(f) sort_free(f);
	regfree(&re);

	return ret;
}

/*
 * Parse -k option value.
 */
static int parse_k(const char* s, key_specs* ks)
{

	int ret = -1;
	bool mef_flags[NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS] = { false, false, false, false, false, false };

	if (s && *s) {

		char* sptr = strchr(s, ',');
		if (sptr) {

			char* pos1, *pos2;
			size_t size1 = sptr - s;

			if (size1 < 1)
				return -1;
			pos1 = (char*) sort_malloc((size1 + 1) * sizeof(char));
			if (!pos1)
				err(2, NULL);

			strncpy(pos1, s, size1);
			pos1[size1] = 0;

			ret = parse_pos(pos1, ks, mef_flags, false);

			sort_free(pos1);
			if (ret < 0)
				return ret;

			pos2 = strdup(sptr + 1);

			ret = parse_pos(pos2, ks, mef_flags, true);

			sort_free(pos2);

		} else {
			ret = parse_pos(s, ks, mef_flags, false);
		}
	}

	return ret;
}

/*
 * Parse POS in +POS -POS option.
 */
static int parse_pos_obs(const char* s, int *nf, int *nc)
{
	int ret = -1;
	regex_t re;
	size_t nmatch = 3;
	regmatch_t pmatch[3];
	char* f = NULL, *c = NULL;
	size_t len;

	*nf = 0;
	*nc = 0;

	const char* sregexp = "^([0-9]+)(\\.[0-9]+)?$";

	if (regcomp(&re, sregexp, REG_EXTENDED) != 0) {
		return -1;
	}

	if (regexec(&re, s, nmatch, pmatch, 0) != 0) {
		goto end;
	}

	if (pmatch[0].rm_eo <= pmatch[0].rm_so) {
		goto end;
	}

	if (pmatch[1].rm_eo <= pmatch[1].rm_so) {
		goto end;
	}

	len = pmatch[1].rm_eo - pmatch[1].rm_so;
	f = (char*) sort_malloc((len + 1) * sizeof(char));
	if (!f)
		err(2, NULL);

	strncpy(f, s + pmatch[1].rm_so, len);
	f[len] = 0;

	*nf = (size_t) strtoul(f, NULL, 10);

	if (pmatch[2].rm_eo > pmatch[2].rm_so) {
		len = pmatch[2].rm_eo - pmatch[2].rm_so - 1;
		c = (char*) sort_malloc((len + 1) * sizeof(char));
		if (!c)
			err(2, NULL);

		strncpy(c, s + pmatch[2].rm_so + 1, len);
		c[len] = 0;

		*nc = (size_t) strtoul(c, NULL, 10);
	}

	ret = 0;

	end:

	if (c)
		sort_free(c);
	if (f)
		sort_free(f);
	regfree(&re);

	return ret;
}

/*
 * "Translate" obsolete +POS1 -POS2 syntax into new -kPOS1,POS2 syntax
 */
void fix_obsolete_keys(int *argc, char** argv)
{
	int i = 0;
	char sopt[129] = "\0";
	for (i = 1; i < *argc; i++) {
		char* arg1 = argv[i];
		if (strlen(arg1) > 1 && arg1[0] == '+') {
			int f1 = 0, c1 = 0;
			if (parse_pos_obs(arg1 + 1, &f1, &c1) < 0) {
				continue;
			} else {
				f1 += 1;
				c1 += 1;
				if (i + 1 < *argc) {
					char* arg2 = argv[i + 1];
					if (strlen(arg2) > 1 && arg2[0] == '-') {
						int f2 = 0, c2 = 0;
						if (parse_pos_obs(arg2 + 1, &f2, &c2) >= 0) {
							if (c2 > 0) {
								f2 += 1;
							}
							sprintf(sopt, "-k%d.%d,%d.%d", f1, c1, f2, c2);
							argv[i] = strdup(sopt);
							{
								int j = 0;
								for (j = i + 1; j + 1 < *argc; j++) {
									argv[j] = argv[j + 1];
								}
								*argc -= 1;
							}
							continue;
						}
					}
				}
				sprintf(sopt, "-k%d.%d", f1, c1);
				argv[i] = strdup(sopt);
			}
		}
	}
}

/*
 * Main function.
 */
int main(int argc, char **argv)
{
	int result = 0;
	int c;
	char *outfile = strdup("-");
	bool mef_flags[NUMBER_OF_MUTUALLY_EXCLUSIVE_FLAGS] = { false, false, false, false, false, false };

	sort_mods *sm = &default_sort_mods_object;

	init_tmp_files();

	set_signal_handler();

	set_free_memory();
	set_locale();
	set_tmpdir();
	set_sort_opts();

#if defined(SUPPORT_PRE_POSIX_KEYS_SYNTAX)
	fix_obsolete_keys(&argc,argv);
#endif

	while (((c = getopt_long(argc, argv, OPTIONS, long_options, NULL)) != -1)) {

		check_mutually_exclusive_flags(c, mef_flags);

		if (!set_sort_modifier(sm, c)) {

			switch (c){
			case 'c':
				sort_opts_vals.cflag = true;
				if (optarg) {
					if (!strcmp(optarg, "diagnose-first")) {
						;
					} else if (!strcmp(optarg, "silent") || !strcmp(optarg, "quiet")) {
						sort_opts_vals.csilentflag = true;
					} else if (*optarg) {
						unknown(optarg);
					}
				}
				break;
			case 'C':
				sort_opts_vals.cflag = true;
				sort_opts_vals.csilentflag = true;
				break;
			case 'k':
			{
				sort_opts_vals.complex_sort = true;
				sort_opts_vals.kflag = true;

				keys_num++;
				keys = (key_specs*) sort_realloc(keys, keys_num * sizeof(key_specs));
				memset(&(keys[keys_num - 1]), 0, sizeof(key_specs));

				if (parse_k(optarg, &(keys[keys_num - 1])) < 0) {
					fprintf(stderr, "invalid -k agrument: %s\n", optarg);
					exit(2);
				}

				break;
			}
			case 'm':
				sort_opts_vals.mflag = true;
				break;
			case 'o':
				outfile = sort_realloc(outfile, sizeof(char) * (strlen(optarg) + 1));
				strlcpy(outfile, optarg, (strlen(optarg) + 1));
				break;
			case 's':
				sort_opts_vals.sflag = true;
				break;
			case 'S':
				available_free_memory = parse_memory_buffer_value(optarg);
				break;
			case 'T':
				tmpdir = strdup(optarg);
				break;
			case 't':
				if (strlen(optarg) > 1) {
					if (strcmp(optarg, "\\0")) {
						fprintf(stderr, "%s: %s\n", strerror(EINVAL), optarg);
						exit(2);
					}
					*optarg = 0;
				}
				sort_opts_vals.tflag = true;
				sort_opts_vals.field_sep = btowc(optarg[0]);
				if (sort_opts_vals.field_sep == WEOF) {
					errno = EINVAL;
					err(2, NULL);
				}
				break;
			case 'u':
				sort_opts_vals.uflag = true;
				/* we need stable sort to choose the correct unique value */
				sort_opts_vals.sflag = true;
				break;
			case 'z':
				sort_opts_vals.zflag = true;
				break;
			case SORT_OPT:
				if (optarg) {
					sort_opts_vals.complex_sort = true;
					if (!strcmp(optarg, "general-numeric")) {
						set_sort_modifier(sm, 'g');
					} else if (!strcmp(optarg, "human-numeric")) {
						set_sort_modifier(sm, 'h');
					} else if (!strcmp(optarg, "numeric")) {
						set_sort_modifier(sm, 'n');
					} else if (!strcmp(optarg, "month")) {
						set_sort_modifier(sm, 'M');
					} else if (!strcmp(optarg, "random")) {
						set_sort_modifier(sm, 'R');
					} else {
						unknown(optarg);
					}
				}
				break;
			case QSORT_OPT:
				sort_opts_vals.sort_method = SORT_QSORT;
				break;
			case MERGESORT_OPT:
				sort_opts_vals.sort_method = SORT_MERGESORT;
				break;
			case HEAPSORT_OPT:
				sort_opts_vals.sort_method = SORT_HEAPSORT;
				break;
			case RADIXSORT_OPT:
				sort_opts_vals.sort_method = SORT_RADIXSORT;
				break;
			case CP_OPT:
				unimplemented("compress-program");
				break;
			case FF_OPT:
				read_fns_from_file0(optarg);
				break;
			case BS_OPT:
			{
				long mof = strtol(optarg, NULL, 10);
				if (mof >= 2) {
					max_open_files = (size_t) mof + 1;
				}
			}
				break;
			case VERSION_OPT:
				printf("%s\n", VERSION);
				exit(0);
				break;
			case DEBUG_OPT:
				debug_sort = true;
				break;
			case HELP_OPT:
				usage(false);
				break;
			default:
				usage(true);
			}
		}
	}

	argc -= optind;
	argv += optind;

#ifndef WITHOUT_NLS
	catalog = catopen("sort", NL_CAT_LOCALE);
#endif

	if (sort_opts_vals.cflag && sort_opts_vals.mflag)
		errx(1, getstr(1), 'm', 'c');

#ifndef WITHOUT_NLS
	catclose(catalog);
#endif

	{
		size_t i = 0;

		if (!keys_num) {
			keys_num = 1;
			keys = (key_specs*) sort_realloc(keys, sizeof(key_specs));
			memset(&(keys[0]), 0, sizeof(key_specs));
			keys[0].c1 = 1;
			keys[0].pos1b = default_sort_mods->bflag;
			keys[0].pos2b = default_sort_mods->bflag;
			memcpy(&(keys[0].sm), default_sort_mods, sizeof(sort_mods));
		}

		for (i = 0; i < keys_num; i++) {
			key_specs *ks = &(keys[i]);
			if (sort_modifier_empty(&(ks->sm)) && !(ks->pos1b) && !(ks->pos2b)) {
				ks->pos1b = sm->bflag;
				ks->pos2b = sm->bflag;
				memcpy(&(ks->sm), sm, sizeof(sort_mods));
			}
		}
	}

	if (argc_from_file0 >= 0) {
		argc = argc_from_file0;
		argv = argv_from_file0;
	}

	if (!sort_opts_vals.cflag && !sort_opts_vals.mflag) {

		sort_list list;
		file_list fl;

		mode_t orig_file_mask = umask(S_IWGRP | S_IWOTH | S_IRGRP | S_IROTH);

		sort_list_init(&list);
		file_list_init(&fl, true);

		if (argc < 1) {
			FILE *file;

			file = openfile("-", "r");
			procfile(file, &list, &fl);
		} else {
			while (argc > 0) {
				FILE *file;

				file = openfile(*argv, "r");
				procfile(file, &list, &fl);
				if (file != stdin)
					fclose(file);
				--argc;
				++argv;
			}
		}

		if (fl.count < 1) {
			umask(orig_file_mask);
			sort_list_to_file(&list, outfile);
		} else {
			if (list.count > 0) {
				char* flast = new_tmp_file_name();
				sort_list_to_file(&list, flast);
				file_list_add(&fl, flast, false);
			}
			umask(orig_file_mask);
			merge_files(&fl, outfile);
		}

		file_list_clean(&fl);

		/* We do not need to clean it at this point */
		/* sort_list_clean(&list); */

	} else if (sort_opts_vals.cflag) {
		if (argc == 0)
			result = (check("-"));
		else
			result = (check(*argv));
	} else if (sort_opts_vals.mflag) {
		file_list fl;
		file_list_init(&fl, false);
		file_list_populate(&fl, argc, argv, true);
		merge_files(&fl, outfile);
		file_list_clean(&fl);
	}

	sort_free(outfile);

	return result;
}
