/*-
 * Copyright (C) 2009 Gabor Kovesdan <gabor@FreeBSD.org>
 * 2012 Oleg Moskalenko oleg.moskalenko@citrix.com
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 */

#include <stdbool.h>
#include <stdio.h>
#include <errno.h>
#include <sysexits.h>
#include <wchar.h>

#define		VERSION	"1.8-FreeBSD"

#define		SORT_DEFAULT	0
#define		SORT_QSORT	1
#define		SORT_MERGESORT	2
#define		SORT_HEAPSORT	3
#define		SORT_RADIXSORT  4

#ifdef WITHOUT_NLS
#define getstr(n)	 nlsstr[n]
#else
#include <nl_types.h>

extern nl_catd catalog;
#define getstr(n)	 catgets(catalog, 1, n, nlsstr[n])
#endif

/* wchar_t is of 4 bytes: */
#define SIZEOF_WCHAR_STRING(LEN) ((LEN)*sizeof(wchar_t))

/*
 * If true, we output some debug information.
 */
extern bool debug_sort;

extern const char *nlsstr[];

/*
 * sort.c
 */

/*
 * This structure holds main sort options which are NOT affecting the sort ordering.
 */
typedef struct _sort_opts
{
	int sort_method;
	bool cflag;
	bool csilentflag;
	bool kflag;
	bool mflag;
	bool sflag;
	bool uflag;
	bool zflag;
	bool tflag;
	wchar_t field_sep;
	bool complex_sort;
} sort_opts;

/*
 * Cmp function
 */

struct _key_value;
typedef struct _key_value key_value;

typedef int (*cmpcoll_t)(key_value *kv1, key_value *kv2);

/*
 * This structure holds "sort modifiers" - options which are affecting the sort ordering.
 */
typedef struct _sort_mods
{
	bool bflag;
	bool dflag;
	bool fflag;
	bool gflag;
	bool iflag;
	bool Mflag;
	bool nflag;
	bool rflag;
	bool Rflag;
	bool Vflag;
	bool hflag;
	cmpcoll_t func;
} sort_mods;

/*
 * Binary "wide" string
 */
typedef struct _bwstring {
	size_t len;
	size_t sz;
	wchar_t str[1];
} bwstring;

/*
 * Sort hint data for -n
 */
typedef struct _n_hint
{
	bool neg1;
	unsigned long long n1;
} n_hint;

/*
 * Sort hint data for -g
 */
typedef struct _g_hint
{
	double d;
	bool nan;
	bool notnum;
} g_hint;

/*
 * Sort hint data for -M
 */
typedef struct _M_hint
{
	int m;
} M_hint;

/*
 * Status of a sort hint object
 */
typedef enum
{
	HS_ERROR=-1,
	HS_UNINITIALIZED=0,
	HS_INITIALIZED=1
} hint_status;

/*
 * Sort hint object
 */
typedef struct _key_hint
{
	hint_status status;
	union {
		n_hint nh;
		g_hint gh;
		M_hint Mh;
	} v;
} key_hint;

/*
 * Key value
 */
struct _key_value
{
	bwstring* k; /* key string */
	key_hint kh; /* key sort hint */
};

/*
 * Set of keys container object.
 */
typedef struct _keys_set
{
	size_t nums;
	key_value* set;
} keys_set;

/*
 * Parsed -k option data
 */
typedef struct _key_specs
{
	size_t f1;
	size_t c1;
	bool pos1b;
	size_t f2;
	size_t c2;
	bool pos2b;
	sort_mods sm;
} key_specs;

extern sort_opts sort_opts_vals;

extern sort_mods* current_sort_mods;
extern sort_mods* const default_sort_mods;

extern key_specs* keys;
extern size_t keys_num;

/*
 * Single entry in sort list.
 */
typedef struct _sort_list_item
{
	bwstring* str;
	keys_set* ks;
} sort_list_item;

/*
 * List of data to be sorted.
 */
typedef struct _sort_list
{
	int count;
	int size;
	unsigned long long memsize;
	sort_list_item** list;
} sort_list;

/*
 * List of files to be sorted
 */
typedef struct _file_list
{
	int count;
	int sz;
	char** fns;
	bool tmp;
} file_list;

/*
 * bwstring inline access functions
 */

static inline wchar_t bws_get(const bwstring *bws, size_t pos)
{
	return bws->str[pos];
}

static inline void bws_put(bwstring *bws, wchar_t c, size_t pos)
{
	bws->str[pos]=c;
}

static inline size_t bwslen(const bwstring* bws)
{
	return bws->len;
}

static inline size_t bwssize(const bwstring *bws)
{
	return bws->sz;
}

static inline wchar_t * bws_getstr(bwstring* bws)
{
	return bws->str;
}

static inline const wchar_t * bws_getcstr(const bwstring* bws)
{
	return bws->str;
}

static inline bool bwsend(const bwstring *bws, const wchar_t *s)
{
	return (bws && s && s>=(bws->str+bws->len));
}

static inline void bws_setlen(bwstring* bws, size_t newlen)
{
	if(bws && newlen!=bws->len && newlen<=bws->sz) {
		bws->len=newlen;
		bws->str[newlen]=L'\0';
	}
}

static inline size_t bws_memsize(bwstring* bws)
{
	return (SIZEOF_WCHAR_STRING(bws->sz)+sizeof(bwstring));
}

/**/

/*
 * Main localised symbols
 */
extern wchar_t symbol_decimal_point;
extern wchar_t symbol_thousands_sep;
extern wchar_t symbol_negative_sign;
extern wchar_t symbol_positive_sign;

/**/

extern unsigned long long free_memory;
extern unsigned long long available_free_memory;
extern const char* tmpdir;

/*
 * Max number of simultaneously open files (including the output file).
 */
extern size_t max_open_files;

/*
 * coll.c
 */

cmpcoll_t get_sort_func(sort_mods *sm);
keys_set* alloc_keys_set(void);
void free_keys_set(keys_set* ks);
void set_key_on_keys_set(keys_set *ks, bwstring *s, size_t ind);

void sort_list_item_init(sort_list_item* si);
size_t sort_list_item_set(sort_list_item* si, bwstring* str);
void sort_list_item_clean(sort_list_item* si);
void sort_list_item_swap(sort_list_item* si1, sort_list_item* si2);

keys_set *preproc(const bwstring *s);
int str_coll(const bwstring *, const bwstring *);
int key_coll(const keys_set *ks1, const keys_set *ks2);
int str_list_coll(const bwstring *str1, sort_list_item **ss2);
int list_coll(sort_list_item **ss1, sort_list_item **ss2);

/*
 * file.c
 */

char* read_file0_line(FILE* f);

void init_tmp_files(void);
void clear_tmp_files(void);
char* new_tmp_file_name(void);

void file_list_init(file_list *fl, bool tmp);
void file_list_add(file_list* fl, char* fn, bool allocate);
void file_list_populate(file_list* fl, int argc, char** argv, bool allocate);
void file_list_clean(file_list *fl);

int check(const char *);
void merge_files(file_list *fl, const char *fn_out);
FILE *openfile(const char *, const char *);
int procfile(FILE *, sort_list *list, file_list *fl);

void sort_list_init(sort_list *l);
void sort_list_add(sort_list *l, bwstring* str);
void sort_list_clean(sort_list *l);
void sort_list_dump(sort_list *l, const char* fn);

void sort_list_to_file(sort_list *list, const char* outfile);

/*
 * mem.c
 */
void * sort_malloc(size_t);
void sort_free(void *ptr);
void * sort_realloc(void *, size_t);

/*
 * bwstring.c
 */

bwstring* bwsalloc(size_t sz);
bwstring* bwsrealloc(bwstring* s, size_t sz);
bwstring* bwsdup(const bwstring *s);
bwstring* bwssdup(const wchar_t *str);
bwstring* bwssbdup(const wchar_t *str, size_t size);
void bwsfree(bwstring* s);
size_t bwscpy(bwstring *dst, const bwstring *src);
bwstring* bwsncpy(bwstring *dst, const bwstring *src, size_t size);
bwstring* bwsnocpy(bwstring *dst, const bwstring *src, size_t offset, size_t size);

int bwscoll_bin(const bwstring* bws1, const bwstring *bws2);

static inline int bwscoll(const bwstring* bws1, const bwstring *bws2)
{
	int res = wcscoll(bws1->str, bws2->str);

	if (res)
		return res;

	return bwscoll_bin(bws1,bws2);
}

int bwscasecoll(const bwstring* bws1, const bwstring *bws2);
int bwsfwrite(const bwstring* bws, FILE* f, bool nl);
bwstring *bwsfgetln(FILE* file, size_t *len, bool zero_ended);

/*
 * radixsort.c
 */

void rxsort(sort_list_item **base, size_t nmemb);
